from typing import Any, List
from http import HTTPStatus

import os
from ...utils import raise_invalid_params
from ...types import BaseEmbeddings

class TextEmbeddings(BaseEmbeddings):
    """支持最新的阿里云模型服务灵积API的文本向量模型"""
    @classmethod
    def allowed_params(cls):
        return {
            "model": "文本嵌入模型的名称",
            "api_key": "API_KEY",
            "output_type": "编码时使用的向量类型",
            **BaseEmbeddings.allowed_params()
        }

    def __init__(self, model: str=None, api_key: str=None, output_type: str="dense", **kwargs):
        raise_invalid_params(kwargs, self.__class__.allowed_params())

        super().__init__(
            model=model or "text-embedding-v3",
            api_key=api_key or os.getenv("DASHSCOPE_API_KEY"),
            dim=kwargs.pop("dim", 1024),
            **kwargs
        )
        self.output_type = output_type

        try:
            import dashscope
        except ImportError:
            raise ImportError(
                "Could not import dashscope package. "
                "Please install it via 'pip install -U dashscope'"
            )

    def query(self, text: str, *args, **kwargs) -> List[float]:
        """
        查询文本向量。
        """
        return self.embed_documents([text], text_type="query", **kwargs)[0]

    def embed_documents(self, texts: List[str], text_type: str="document", **kwargs) -> List[List[float]]:
        """
        编码文本向量。
        """
        import dashscope
        dashscope.api_key = self.api_key
        from dashscope import TextEmbedding

        response = TextEmbedding.call(
            model=self.model,
            input=texts,
            text_type=text_type,
            dimension=self.dim,
            output_type=self.output_type,
            **kwargs
        )
        if response.status_code != HTTPStatus.OK:
            raise Exception('Request id: %s, Status code: %s, error code: %s, error message: %s' % (
                response.request_id, response.status_code,
                response.code, response.message
            ))

        chunks = []
        for chunk in response.output['embeddings']:
            chunks.append(chunk['embedding'])
            self.dim = len(chunk['embedding'])

        return chunks




import os
from typing import Union

from .....config import get_env
from .....utils import extract_segments, minify_text, filter_kwargs, raise_invalid_params
from .....io import EventBlock
from ...selector import Selector, End
from ...prompt_template import PromptTemplate
from ...message import Messages
from ..base import BaseAgent
from ..chat import ChatAgent
from .base import FlowAgent

def save_faq(thread_id: str, chat_learn_folder: str, knowledge: str, question: str="", metadata: dict={}):
    if not thread_id or not knowledge:
        return

    if not os.path.exists(chat_learn_folder):
        os.makedirs(chat_learn_folder)

    metadata = f'<!-- @metadata {str(metadata) if metadata else ""} -->\n'
    q = f"**Question**\n{question}\n\n"
    k = f"**Knowledge**\n{knowledge}"
    text = (metadata + q + k) or ""

    with open(os.path.join(chat_learn_folder, f"{thread_id}.md"), "w", encoding="utf-8") as f:
        f.write(text)
    return text

class ChatLearn(FlowAgent):
    """
    ChatLearn，用于从对话中学习知识。
    """
    @classmethod
    def allowed_params(cls):
        return {
            "scribe": "负责笔记的ChatAgent",
            "scribe_template": "scribe 所使用的 PromptTemplate, 默认为 PromptTemplate('FLOW/Scribe')",
            **FlowAgent.allowed_params(),
        }

    def __init__(
        self,
        scribe: ChatAgent,
        scribe_template: str=None,
        **kwargs
    ):
        raise_invalid_params(kwargs, self.allowed_params())

        if not isinstance(scribe, ChatAgent):
            raise ValueError("scribe 必须是 ChatAgent 的子类")

        scribe_template = scribe_template or PromptTemplate("FLOW/Scribe")
        self.scribe_template = scribe_template

        scribe.reset_init_memory(scribe_template)
        self.scribe = scribe

        def fetch_faq(agent: BaseAgent):
            final_output_text = agent.last_output
            questions = extract_segments(final_output_text, ('<question>', '</question>'))
            knowledges = extract_segments(final_output_text, ('<knowledge>', '</knowledge>'))
            metadata = {"class": scribe.__class__.__name__, "name": scribe.name, "thread_id": scribe.thread_id}

            # 保存 Q/K 语料
            for i, knowledge in enumerate(knowledges):
                q = questions[i] if i < len(questions) else ""
                text = save_faq(scribe.thread_id, scribe.chat_learn_folder, knowledge, q, metadata)
                yield EventBlock("faq", f"保存知识到[{scribe.thread_id}]：{minify_text(q)} -> {minify_text(knowledge)}")
                scribe.clear()
                if scribe.default_vdb:
                    scribe.default_vdb.load_text(text, source=scribe.thread_id)

        def should_fetch():
            if '<knowledge>' in scribe.last_output:
                return "Fetch_FAQ"
            else:
                return "__END__"

        super().__init__(
            {"Scribe": scribe},
            Selector(should_fetch),
            {"Fetch_FAQ": fetch_faq},
            End(),
            **filter_kwargs(kwargs, self.allowed_params())
        )

        self.description = f"我擅长提炼新知识，所有涉及到总结、提炼新知识的地方都可以召唤我来完成任务"
        self.tool_params = {
            "prompt": "请给出你要总结的内容，并在结尾附上一句`请帮我总结其中的新知识`"
        }

    def call(self, prompt: Union[str, ChatAgent], **kwargs):
        return super().call(prompt, **kwargs)

    def begin_call(self, args):
        from ..chat import ChatAgent
        if isinstance(args[0], ChatAgent):
            self.task = args[0].task
            memory = Messages(args[0].memory)
            return [f'请帮我从这段对话过程中提取知识：\n\n{str(memory.to_list(style="text"))}']
        else:
            self.task = args[0] if args else None
            return args




from abc import ABC, abstractmethod
from typing import List
import time
import random

class ThreadIDGenerator:
    def __init__(self, counter: int=0):
        self.counter = counter

    def create_id(self, last_count: str=None):
        if last_count:
            self.counter = int(last_count)
        while True:
            timestamp = str(int(time.time()))[-6:]
            random_number = f'{random.randint(0, 9999):04}'
            counter_str = f'{self.counter:04}'
            yield f'{timestamp}-{random_number}-{counter_str}'
            self.counter = 0 if self.counter == 9999 else self.counter + 1

thread_id_gen = ThreadIDGenerator()


class BaseHistory(ABC):
    def __init__(self, agent_class: str="CHAT_AGENT", agent_name: str="default"):
        self.reset_init(agent_class, agent_name)

    def reset_init(self, agent_class: str, agent_name: str):
        self.agent_class = agent_class
        self.agent_name = agent_name

    @abstractmethod
    def list_threads(self):
        """列举所有记忆线"""
        pass

    @property
    def last_thread_id(self):
        all_thread_ids = self.list_threads()
        return all_thread_ids[-1] if all_thread_ids else None

    @abstractmethod
    def last_thread_id_count(self):
        """获取最近一轮对话的线程 ID"""
        pass

    @abstractmethod
    def save_memory(self, thread_id: str, memory: List[dict]):
        """根据 thread_id 保存记忆"""
        pass

    @abstractmethod
    def load_memory(self, thread_id: str):
        """根据 thread_id 加载记忆"""
        pass



from typing import List, Dict, Any
from ......io import EventBlock, NewLineBlock
from ....message import Messages
from ...base import BaseAgent
from .base import BaseToolCalling
import re
import json

class Plans(BaseToolCalling):
    def __init__(self, steps: Dict[str, Dict[str, Any]] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def completed_work(self):
        """
        返回所有步骤的完成情况。
        """
        completed_work = []
        for step in self.steps:
            desc = f'Step{step.get("index", 0)}: {step.get("description", "No description")}'
            e = f'{step.get("eid", "#E0")} = {step.get("name", "")}[{step.get("arguments", "")}]'
            result = step.get("result", None)
            if result:
                result = f'{step.get("name", "")} Value = {result}'
            completed_work.append(f'{desc} \n{e} \n{result or ""}')
        return "\n".join(completed_work)

    def extract_tools_call(self, text: str) -> Dict[str, Dict[str, Any]]:
        # 正则表达式解析文本
        # Regex to match expressions of the form Step1: ... #E... = ...[{"p1":"v1","p2":"v2"}]
        pattern_full = r"Step(\d+):\s*(.+?)\s*#(E\d+)\s*=\s*(\w+)(?:[\(\[](\{.*?\})[\)\]])?"
        pattern_partial = r"Step(\d+):\s*(.*)(#E\d+)\s*"
        steps = []
        
        for line in text.splitlines():
            if not line:
                continue

            match_full = re.search(pattern_full, line)
            
            if match_full:
                index = match_full.group(1).strip()
                description = match_full.group(2).strip()
                eid = match_full.group(3).strip()
                function_name = match_full.group(4).strip()
                arguments = match_full.group(5).strip() if match_full.group(5) else ""
                
                # 确保 match_full 的关键字段不为空
                if eid and function_name:
                    step = {
                        "index": f"{index}",
                        "eid": eid,
                        "description": description,
                        "name": function_name,
                        "arguments": arguments,
                        "result": ""
                    }
                    steps.append(step)
                    continue

            # 如果 match_full 不成功或关键字段为空，尝试 match_partial
            match_partial = re.search(pattern_partial, line)
            if match_partial:
                index = match_partial.group(1).strip()
                eid = f'#E{index}'
                description = match_partial.group(2).strip() + f" {eid}"
                function_name = ""
                arguments = ""
            else:
                continue

            step = {
                "index": f"{index}",
                "eid": eid,
                "description": description,
                "name": function_name,
                "arguments": arguments,
                "result": None
            }
            steps.append(step)

        self.steps.extend(steps)
        return steps

    def handle(self, steps: List[Any], **kwargs):
        """
        处理单个计划的调用。
        """
        pre_build_vars = {}

        for step in steps:
            arguments = step.get("arguments", "")
            # 查找并替换占位符
            placeholders = re.findall(r"#E\d+", arguments)
            for placeholder in placeholders:
                if placeholder in pre_build_vars:
                    # 假设每个步骤的结果是一个字符串
                    result = pre_build_vars[placeholder]
                    if result:
                        # 情况2: "#E{n}" 的结构，有双引号包围
                        if re.fullmatch(rf'"{placeholder}"', arguments):
                            arguments = arguments.replace(f'"{placeholder}"', f'"{result}"')
                        # 情况3: "otherword#E{n}otherword" 的字结构，有双引号包围
                        elif re.search(rf'"[^"]*{placeholder}[^"]*"', arguments):
                            arguments = arguments.replace(placeholder, result)
                        # 情况4: #E{n}，但没有双引号包围
                        elif re.fullmatch(rf'{placeholder}', arguments):
                            arguments = arguments.replace(placeholder, json.dumps(result, ensure_ascii=False))

            # 确保 arguments 是有效的 JSON
            if arguments:  # 仅在 arguments 非空时尝试解析
                try:
                    parsed_arguments = json.loads(arguments)
                    arguments = json.dumps(parsed_arguments, ensure_ascii=False)

                    tool_to_exec = {
                        "function": {
                            "name": step.get("name"),
                            "arguments": arguments
                        }
                    }
                    for block in self.execute_tool(tool_to_exec):
                        if block.block_type == "final_tool_resp":
                            # 将结果存储到 pre_build_vars 中
                            pre_build_vars[step["eid"]] = block.text
                            step["result"] = block.text.strip()
                        yield block

                except json.JSONDecodeError as e:
                    print(f"JSONDecodeError: {e} in arguments: {arguments}")

    async def async_handle(self, final_tool_call):
        async for block in self.async_execute_tool(final_tool_call):
            yield block

    @property
    def tools_desc(self) -> str:
        """
        将工具描述转换为 Plan 模式。
        """
        desc = []
        for index, t in enumerate(self.tools_to_exec):
            tool_desc = []
            for k, v in t.parameters["properties"].items():
                tool_desc.append(f"{k}的类型为{v['type']},(目的是){v['description']}")
            desc.append(f'({index+1}){t.name}[{",".join([p for p in t.parameters["properties"]])}]: {t.description}。 {";".join(tool_desc)}')
        return "\n".join(desc)



from typing import List, Union

from .....utils import raise_invalid_params, minify_text
from .....io import EventBlock
from ....document import Document
from ...selector import Selector
from ...prompt_template import PromptTemplate
from ..base import BaseAgent
from ..chat import ChatAgent
from ..flow import FlowAgent
from .markdown import Markdown

import copy
import asyncio

class FromOutline(BaseAgent):
    """
    实现扩写：从输出结果提炼大纲，然后生成内容。
    """
    @classmethod
    def allowed_params(cls):
        return {
            "writer": "扩写智能体",
            "prev_k": "前 k 个字符",
            "next_k": "后 k 个字符"
        }

    def __init__(self, writer: Union[ChatAgent, FlowAgent]=None, template_id: str=None, prev_k:int=1000, next_k:int=500, **kwargs):
        raise_invalid_params(kwargs, self.__class__.allowed_params())

        if not isinstance(writer, (ChatAgent, FlowAgent, Selector, BaseAgent)):
            raise ValueError("扩写智能体 writer 必须是 ChatAgent 或 FlowAgent 或 Selector 或 BaseAgent 实例")

        super().__init__(**kwargs)
        self.writer = writer
        self.template_id = template_id or "WRITER/FromOutline"

        self.prev_k = prev_k
        self.next_k = next_k

        # 执行 call 调用时提供
        self.outline_text = None
        self.markdown = None

        # 执行 call 调用时生成
        self.segments = []

    @property
    def last_output(self):
        """
        从 Markdown 生成当前的文本输出内容。
        如果当前有提纲，则将提纲中的内容替换为实际内容。
        """
        if self.segments:
            md = copy.deepcopy(self.markdown)
            for (thread_id, doc_id, from_outline_text) in self.segments:
                md.replace_documents(doc_id, doc_id, [Document(from_outline_text + "\n")])
            return md.text
        else:
            return self.outline_text

    def fetch_outline(self):
        """
        提取大纲。
        """
        self.markdown = Markdown(self.outline_text)
        return self.markdown.get_outline() if self.markdown else []

    def call(self, outline: Union[str, BaseAgent], *args, **kwargs):
        """
        执行扩写。
        """

        if isinstance(outline, str):
            self.outline_text = outline
        elif isinstance(outline, BaseAgent):
            self.outline_text = outline.last_output
        else:
            raise ValueError("outline 必须是字符串或 BaseAgent 实例")

        # 提取大纲
        outline_docs = self.fetch_outline()

        if outline_docs:
            for doc in outline_docs:
                outline_id = doc.meta['id']
                (draft, outline) = self.markdown.fetch_outline_task(doc, prev_k=self.prev_k, next_k=self.next_k)

                info = f"执行扩写任务 <{outline_id}>：\n{minify_text(outline)}"
                yield EventBlock("agent", info)

                self.bind_consumer(
                    self.writer,
                    binding_map={
                        "task": lambda: outline,
                        "outline": lambda: f'```markdown\n{outline}\n```',
                        "draft": lambda: f'```markdown\n{draft}\n```'
                    }
                )

                messages_for_chat = [
                    (
                        'system',
                        PromptTemplate(
                            self.template_id,
                            binding_map={
                                "outline": lambda: f'```markdown\n{outline}\n```',
                                "draft": lambda: f'```markdown\n{draft}\n```'
                            }).format()
                    ),
                    (
                        'user',
                        "请开始扩写"
                    )
                ]

                if isinstance(self.writer, FlowAgent):
                    for a in self.writer.agents:
                        if isinstance(a, ChatAgent):
                            a.clear()
                    input_text = PromptTemplate(self.template_id).format({
                        "outline": lambda: f'```markdown\n{outline}\n```',
                        "draft": lambda: f'```markdown\n{draft}\n```'
                    })
                    yield from self.writer.call(input_text)
                elif isinstance(self.writer, ChatAgent):
                    if self.writer.init_memory:
                        yield from self.writer.call(outline, new_chat=True)
                    else:
                        yield from self.writer.call(messages_for_chat)
                elif isinstance(self.writer, Selector):
                    self.writer.select()
                    if isinstance(self.writer.selected, ChatAgent): 
                        if not self.writer.selected.init_memory:
                            yield from self.writer.selected.call(messages_for_chat)
                        else:
                            yield from self.writer.selected.call(outline)
                    else:
                        yield from self.writer.selected.call(outline)
                else:
                    yield from self.writer.call(outline, *args, **kwargs)

                self.segments.append(
                    (
                        self.writer.selected.calling_id,
                        outline_id,
                        self.writer.selected.last_output
                    )
                )

        else:
            yield EventBlock("info", f"没有提纲可供扩写")

    async def async_call(self, outline: Union[str, BaseAgent], *args, **kwargs):
        """
        执行扩写。

        注意，该异步版本使用使用 asyncio.gather 并行处理异步生成器时，
        所有的结果会在所有任务完成后一次性返回。
        这是因为 asyncio.gather 会等待所有协程完成，然后返回结果。 
        """

        if isinstance(outline, str):
            self.outline_text = outline
        elif isinstance(outline, BaseAgent):
            self.outline_text = outline.last_output
        else:
            raise ValueError("outline 必须是字符串或 BaseAgent 实例")

        # 提取大纲
        outline_docs = self.fetch_outline()

        async def process_doc(doc):
            outline_id = doc.meta['id']
            (draft, outline) = self.markdown.fetch_outline_task(doc, prev_k=self.prev_k, next_k=self.next_k)

            info = f"执行扩写任务 <{outline_id}>：\n{minify_text(outline)}"
            yield EventBlock("agent", info)

            if isinstance(self.writer, FlowAgent):
                input_text = PromptTemplate(self.template_id).format({
                    "outline": lambda: f'```markdown\n{outline}\n```',
                    "draft": lambda: f'```markdown\n{draft}\n```'
                })
                async for block in self.writer.async_call(input_text):
                    yield block

            elif isinstance(self.writer, ChatAgent):
                async for block in self.writer.async_call([
                    (
                        'system',
                        PromptTemplate(
                            self.template_id,
                            binding_map={
                                "outline": lambda: f'```markdown\n{outline}\n```',
                                "draft": lambda: f'```markdown\n{draft}\n```'
                            }).format()
                    ),
                    (
                        'user',
                        "请开始扩写"
                    )
                ]):
                    yield block

            # 确保在处理完所有块后再更新 segments
            self.segments.append(
                (
                    self.writer.thread_id,
                    outline_id,
                    self.writer.last_output
                )
            )

        async def gather_docs():
            if outline_docs:
                tasks = [process_doc(doc) for doc in outline_docs]
                results = await asyncio.gather(*[self._consume_async_gen(process_doc(doc)) for doc in outline_docs])
                for result in results:
                    for block in result:
                        yield block
            else:
                yield EventBlock("info", f"没有提纲可供扩写")

        async for block in gather_docs():
            yield block

    async def _consume_async_gen(self, agen):
        """
        Helper function to consume an async generator and return a list of results.
        """
        results = []
        async for item in agen:
            results.append(item)
        return results




from typing import List, Dict, Any
from .base import BaseToolCalling
from ......io import EventBlock, NewLineBlock
from ....message import Messages
import json

class ToolCall(BaseToolCalling):

    def extract_tools_call(self, text: str) -> List[Dict[str, Any]]:
        start_marker = "<tool_call>"
        end_marker = "</tool_call>"
        start = text.find(start_marker)
        steps = []
        while start != -1:
            end = text.find(end_marker, start)
            if end != -1:
                tool_call_json = text[start + len(start_marker):end]
                try:
                    tool_call = json.loads(tool_call_json)
                    steps.append({
                        "function": {
                            "name": tool_call.get("name"),
                            "arguments": json.dumps(tool_call.get("arguments", "[]"), ensure_ascii=False)
                        }
                    })
                except json.JSONDecodeError:
                    pass
                start = text.find(start_marker, end)
            else:
                break

        for index, step in enumerate(steps):
            name = step["function"].get("name", "UNKNOWN_FUNCTION")
            arguments = step["function"].get("arguments", "")
            self.steps.append({
                "index": index + 1,
                "eid": f"#E{index + 1}",
                "description": f"调用{name}工具",
                "name": name,
                "arguments": arguments,
                "result": None
            })

        return steps

    def handle(self, steps: List[Any], short_term_memory: Messages, long_term_memory: Messages):
        for index, tool_call in enumerate(steps):
            # 由于工具是从文本中提取的，第一个工具在文本中已经作为AI回复，所以从第二个开始构造记忆
            if index > 0:
                new_task = f'请继续: {json.dumps(tool_call, ensure_ascii=False)}'
                tool_call_message = [
                    {
                        "role": "assistant",
                        "content": new_task
                    }
                ]
                short_term_memory.extend(tool_call_message)
                long_term_memory.extend(tool_call_message)

            # 执行工具
            for block in self.execute_tool(tool_call):
                if isinstance(block, EventBlock) and block.block_type == "final_tool_resp":
                    tool_resp = block.text
                    tool_resp_message = [
                        {
                            "role": "user",
                            "content": f'<tool_resp>{tool_resp}</tool_resp>'
                        }
                    ]
                    short_term_memory.extend(tool_resp_message)
                    long_term_memory.extend(tool_resp_message)
                    # 补充结果到 steps
                    self.steps[index]["result"] = tool_resp.strip()
                yield block

    async def async_handle(self, steps: List[Any], short_term_memory: Messages, long_term_memory: Messages):
        for index, tool_call in enumerate(steps):
            # 由于工具是从文本中提取的，第一个工具在文本中已经作为AI回复，所以从第二个开始构造记忆
            if index > 0:
                new_task = f'请继续: {json.dumps(tool_call, ensure_ascii=False)}'
                tool_call_message = [
                    {
                        "role": "assistant",
                        "content": new_task
                    }
                ]
                short_term_memory.extend(tool_call_message)
                long_term_memory.extend(tool_call_message)

            # 执行工具
            async for block in self.async_execute_tool(tool_call):
                if isinstance(block, EventBlock) and block.block_type == "final_tool_resp":
                    tool_resp = block.text
                    tool_resp_message = [
                        {
                            "role": "user",
                            "content": f'<tool_resp>{tool_resp}</tool_resp>'
                        }
                    ]
                    short_term_memory.extend(tool_resp_message)
                    long_term_memory.extend(tool_resp_message)
                    # 补充结果到 steps
                    self.steps[index]["result"] = tool_resp.strip()
                yield block



from ..community.zhipu import ZhipuEmbeddings
from ..community.dashscope import TextEmbeddings
from ..community.openai import OpenAIEmbeddings
from ..community.hugging_face import HuggingFaceEmbeddings

from ..core.runnable import Retriever, MarkMeta
from ..community.faiss import FaissDB
from ..community.dashscope.reranker import DashScopeReranker

__all__ = [
    "ZhipuEmbeddings",
    "TextEmbeddings",
    "OpenAIEmbeddings",
    "HuggingFaceEmbeddings",
    "FaissDB",
    "Retriever",
    "MarkMeta",
    "DashScopeReranker",
]



from typing import Union, List, Callable

from .....utils import filter_kwargs, raise_invalid_params
from .....io import EventBlock
from ...selector import Selector
from ...prompt_template import PromptTemplate
from ..base import BaseAgent
from ..chat import ChatAgent
from .base import FlowAgent

class CoT(FlowAgent):
    """
    CoT 是 ReAct 的简化版，提供了一种没有工具的反思过程，有利于单步思考时避免不必要的工具回调。
    """
    @classmethod
    def allowed_params(cls):
        return {
            "planner": "用于推理的ChatAgent, 你应当在其中指定可用的工具",
            "planner_template": "用于生成计划的PromptTemplate, 默认为 PromptTemplate('FLOW/CoT/Planner')",
            **FlowAgent.allowed_params(),
        }

    def __init__(
        self,
        planner: BaseAgent,
        planner_template: str=None,
        **kwargs
    ):
        raise_invalid_params(kwargs, self.allowed_params())

        if not isinstance(planner, ChatAgent):
            raise ValueError("planner 必须是 ChatAgent 的子类")

        planner_template = planner_template or PromptTemplate("FLOW/CoT/Planner")
        self.planner_template = planner_template

        planner.reset_init_memory(planner_template)
        self.planner = planner
        self.completed_work = []

        def observe_func(agent: BaseAgent):
            output = agent.last_output

            if output:
                self.completed_work.append(output)
            else:
                if not planner.final_answer:
                    yield EventBlock("warn", f"观察结果中没有 final_answer，可能导致无法结束")

            if agent.last_output:
                observation = f'\n**观察**\n上面的行动结果为:\n{agent.last_output}\n'

                self.completed_work.append(observation)
                planner_template.bind_provider({
                    "completed_work": "\n".join(self.completed_work)
                })

            yield EventBlock("final_text", self.task)

        def should_continue(vars, runs):
            if planner.final_answer:
                return "__END__"
            else:
                planner.clear()
                return "planner"

        super().__init__(
            {"planner": planner},
            {"observer": observe_func},
            Selector(condition=should_continue, name="should_continue"),
            **filter_kwargs(kwargs, self.allowed_params())
        )
    
    def begin_call(self, args):
        self.planner.clear()
        self.completed_work.clear()
        self.planner_template.bind_provider({"completed_work": ""})

        return super().begin_call(args)

    def end_call(self):
        self._last_output = self.planner.final_answer



import os
import json

from typing import Union, List

from ...config import get_env
from .base import BaseHistory

class LocalFileHistory(BaseHistory):
    """基于文件的记忆管理"""

    def __init__(self, directory: str = None, **kwargs):
        super().__init__(**kwargs)
        self.directory = directory or get_env("ILLUFLY_LOCAL_FILE_MEMORY")

    def last_thread_id_count(self):
        all_thread_ids = self.list_threads()
        if all_thread_ids:
            ids = all_thread_ids[-1].split("-")
            return int(ids[-1]) + 1
        else:
            return 0

    # 列举所有记忆线
    def list_threads(self):
        memory_dir = self._get_history_dir()
        if not os.path.exists(memory_dir):
            return []
        file_list = [os.path.basename(file) for file in os.listdir(memory_dir) if file.endswith(".json") and not file.startswith(".")]
        thread_ids = [file.replace(".json", "") for file in file_list]

        def thread_id_key(thread_id):
            ids = thread_id.split("-")
            return f'{ids[0]}-{ids[-1]}'

        return sorted(thread_ids, key=thread_id_key)

    def save_memory(self, thread_id: str, memory: List[dict]):
        path = self._get_history_file_path(thread_id)
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, "w", encoding="utf-8") as f:
            json.dump(memory, f, ensure_ascii=False)

    def load_memory(self, thread_id: Union[str, int]=None):
        """
        加载记忆。

        如果 thread_id 是字符串，则直接加载指定线程的记忆；
        如果 thread_id 是整数，则将其当作索引，例如 thread_id=-1 表示加载最近一轮对话的记忆。
        """
        path = None
        _thread_id = thread_id
        if isinstance(thread_id, str):
            path = self._get_history_file_path(thread_id)
        elif isinstance(thread_id, int):
            if self.list_threads():
                _thread_id = self.list_threads()[thread_id]
                path = self._get_history_file_path(_thread_id)

        if path and os.path.exists(path):
            with open(path, "r", encoding="utf-8") as f:
                return _thread_id, json.load(f)

        return _thread_id, []

    def _get_history_dir(self):
        return os.path.join(self.directory, self.agent_class.upper(), self.agent_name)

    def _get_history_file_path(self, thread_id: str):
        if thread_id:
            return os.path.join(
                self._get_history_dir(),
                f"{thread_id}.json"
            )
        else:
            raise ValueError("thread_id MUST not be None")



import json
import os
import asyncio
import aiohttp
import time

from typing import Union, List, Optional, Dict, Any
from http import HTTPStatus
from urllib.parse import urlparse, unquote
from pathlib import PurePosixPath
import requests

from ...types import BaseAgent
from ...utils import raise_invalid_params
from ..http import (
    EventBlock,

    get_headers,
    validate_output_path,

    send_request,
    check_task_status,
    save_resource,

    async_send_request,
    async_check_task_status,
    async_save_resource,

    DASHSCOPE_BASE_URL,
    confirm_upload_file
)

class Text2ImageWanx(BaseAgent):
    """    
    :parameters.size: 生成图像的分辨率，目前仅支持：
        - 1024*1024 默认
        - 720*1280
        - 1280*720

    :parameters.ref_mode: 垫图（参考图）生图使用的生成方式，可选值为
        - repaint 参考内容，默认
        - refonly 参考风格

    [详细调用参数可参考通义万相文生图 API](https://help.aliyun.com/zh/dashscope/developer-reference/api-details-9)
    """
    @classmethod
    def allowed_params(cls):
        return {
            "model": "模型名称",
            "api_key": "API_KEY",
            **BaseAgent.allowed_params()
        }

    def __init__(self, model: str=None, api_key: str=None, **kwargs):
        raise_invalid_params(kwargs, self.__class__.allowed_params())

        super().__init__(threads_group="WANX", **kwargs)

        self.description = "我擅长根据你的文字提示描述生成图片，你必须在 prompt 中详细描述生成要求，你必须展开描述，比如镜头、光线、细部等。"
        self.tool_params = {
            "prompt": "请尽量详细描述生成要求的细节",
            "ref_img": "生成图片时所使用的参考图，可使用资源中提供的图片文件名称",
            "image_count": "生成图片的数量",
            "image_style": "生成图片的风格可以是: 摄影, 人像写真, 3D卡通, 动画, 油画, 水彩, 素描, 中国画, 扁平插画, 默认",
            "output": "指定生成图片的名称，应当包括扩展名"
        }

        self.model = model or "wanx-v1"
        self.api_key = api_key or os.getenv("DASHSCOPE_API_KEY")
    
    def get_style(self, style: str):
        if style == "摄影":
            return "<photography>"
        elif style == "人像写真":
            return "<portrait>"
        elif style == "3D卡通":
            return "<3d cartoon>"
        elif style == "动画":
            return "<anime>"
        elif style == "油画":
            return "<oil painting>"
        elif style == "水彩":
            return "<watercolor>"
        elif style == "素描":
            return "<sketch>"
        elif style == "中国画":
            return "<chinese painting>"
        elif style == "扁平插画":
            return "<flat illustration>"
        else:
            return "<auto>"

    def confirm_content_url(self, input: Dict[str, Any], key: str, tail: str="_url") -> str:
        url = input.get(f"{key}{tail}", None)
        file = input.get(key, None)
        return confirm_upload_file(self.model, file, self.api_key) if file else url

    def get_headers(self):
        return {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.api_key}",
            "X-DashScope-OssResourceResolve": "enable",
            "X-DashScope-Async": "enable"
        }

    def get_data(self, input: Dict[str, Any], parameters: Dict[str, Any]=None):
        ref_img = input.get("ref_img", None)
        ref_img_url = confirm_upload_file(self.model, ref_img, self.api_key) if ref_img else None
        input.update({"ref_img": ref_img_url})
        return {
            "model": self.model,
            "input": {
                "prompt": input.get("prompt", "")[:500],
                "negative_prompt": input.get("negative_prompt", "")[:500] or None,
                "ref_img": self.confirm_content_url(input, "ref_img", "") or None
            },
            "parameters": parameters or {}
        }
    
    def parse_result(self, status_result, output):
        n = len(output or [])
        if status_result['output']['task_status'] == 'SUCCEEDED':
            results = []
            if "results" in status_result['output']:
                results = status_result['output']['results']
            elif "result_url" in status_result['output']:
                result_url = status_result['output']['result_url']
                if isinstance(result_url, list):
                    results = [{"url": r} for r in result_url]
                elif isinstance(result_url, str):
                    results = [{"url": result_url}]
                else:
                    raise Exception(f"output not be list or str, but got {type(result_url)}: output={result_url}")
            else:
                yield EventBlock("warn", "生成失败")

            if results:
                for result_index, result in enumerate(results):
                    url = result['url']
                    yield EventBlock("image_url", url)
                    if not output or result_index >= len(output):
                        parsed_url = urlparse(url)
                        filename = os.path.basename(parsed_url.path)
                        output_path = f"{filename.rsplit('.', 1)[0]}.{filename.rsplit('.', 1)[1]}"
                    else:
                        output_path = output[result_index]
                    yield from save_resource(url, output_path)

                if 'usage' in status_result:
                    yield EventBlock("usage", json.dumps(status_result["usage"], ensure_ascii=False))

    def parse_resp(self, resp):
        yield EventBlock("info", f'{json.dumps(resp, ensure_ascii=False)}')
        # 解析提交结果
        if "output" in resp:
            task_id = resp['output']['task_id']        
            yield EventBlock("info", f'{task_id}: {resp["output"]["task_status"]}')
            if "usage" in resp:
                yield EventBlock("usage", json.dumps(resp["usage"]))
    
    @property
    def aigc_base_url(self):
        return f"{DASHSCOPE_BASE_URL}/services/aigc/text2image/image-synthesis"
    
    def call(
        self, 
        # input 和 parameters 为官方API所需要的详细惨受
        input: Dict[str, Any]={},
        parameters: Optional[Dict[str, Any]]={},
        # 以下为方便工具使用的简易参数
        output: Optional[Union[str, List[str]]] = None,
        prompt: str=None,
        ref_img: str=None,
        image_count: int=1,
        image_style: str="auto",
        **kwargs
    ):
        self._last_output = []
        if prompt:
            input.update({"prompt": prompt})
        if ref_img:
            input.update({"ref_img": ref_img})
        if image_count:
            parameters.update({"n": image_count})
        if image_style:
            parameters.update({"style": self.get_style(image_style)})

        parameters = parameters or {}
        if isinstance(output, str):
            output = [output]

        # 异步提交生成请求
        headers = self.get_headers()
        data = self.get_data(input, parameters)
        resp = send_request(self.aigc_base_url, headers, data)


        # 等待异步生成结果
        yield from self.parse_resp(resp)
        status_result = {}
        if "output" in resp:
            yield from check_task_status(status_result, resp['output']['task_id'], headers)

        # 解析最终结果
        if "output" in status_result:
            yield from self.parse_result(status_result, output)

    async def async_call(
        self, 
        input: Dict[str, Any]=None,
        parameters: Optional[Dict[str, Any]]=None,
        output: Optional[Union[str, List[str]]] = None,
        **kwargs
    ):
        self._last_output = []
        if isinstance(output, str):
            output = [output]

        # 异步提交生成请求
        headers = self.get_headers()
        data = self.get_data(input, parameters)
        resp = await async_send_request(self.aigc_base_url, headers, data)

        # 解析异步提交响应
        for block in self.parse_resp(resp):
            yield block

        # 等待生成结果
        status_result = {}
        async for block in async_check_task_status(status_result, resp['output']['task_id'], headers):
            yield block
        for block in self.parse_result(status_result, output):
            yield block

class CosplayWanx(Text2ImageWanx):
    """
    通义万相-Cosplay动漫人物生成通过输入人像图片和卡通形象图片，可快速生成人物卡通写真。目前支持3D卡通形象风格。
    一张人像照片 + 一张卡通风格 = 一张卡通写真
    详细调用参数可参考通义万相 Cosplay API。

    :input.model_index: 生成风格。取值范围：
        - 1: 3D卡通形象(3dcartoon)
    """
    def __init__(self, model: str=None, **kwargs):
        super().__init__(model=(model or "wanx-style-cosplay-v1"), **kwargs)

    @property
    def aigc_base_url(self):
        return f"{DASHSCOPE_BASE_URL}/services/aigc/image-generation/generation"

    def get_data(self, input: Dict[str, Any], parameters: Dict[str, Any]=None):
        return {
            "model": self.model,
            "input": {
                "face_image_url": self.confirm_content_url(input, "face_image", "_url"),
                "template_image_url": self.confirm_content_url(input, "template_image", "_url"),
                "model_index": input.get("model_index", 1)
            }
        }

class StyleRepaintWanx(Text2ImageWanx):
    """
    通义万相-人像风格重绘可以将输入的人物图像进行多种风格化的重绘生成，
    使新生成的图像在兼顾原始人物相貌的同时，带来不同风格的绘画效果。
    当前支持预置重绘风格和客户上传风格参考图，
    预置重绘风格有复古漫画、3D童话、二次元、小清新、未来科技、国画古风、将军百战等。

    :input.style_index: 风格索引，取值范围为1-10，默认为-1。
        想要生成的风格化类型索引：
            -1 参考上传图像风格
            0 复古漫画
            1 3D童话
            2 二次元
            3 小清新
            4 未来科技
            5 国画古风
            6 将军百战
            7 炫彩卡通
            8 清雅国风
            9 喜迎新年
    """
    def __init__(self, model: str=None, **kwargs):
        super().__init__(model=(model or "wanx-style-repaint-v1"), **kwargs)

    @property
    def aigc_base_url(self):
        return f"{DASHSCOPE_BASE_URL}/services/aigc/image-generation/generation"

    def get_data(self, input: Dict[str, Any], parameters: Dict[str, Any]=None):
        return {
            "model": self.model,
            "input": {
                "image_url": self.confirm_content_url(input, "image", "_url"),
                "style_ref_url": self.confirm_content_url(input, "style_ref", "_url"),
                "style_index": input.get("style_index", 1)
            }
        }

class RepaintWanx(Text2ImageWanx):
    """
    图像局部重绘是根据用户提供原始图片和局部涂抹图片，再加上文字描述，
    然后在涂抹区域生成文字描述的内容，涂抹区域外没有变化。

    :parameters.style: 涂抹修改区域的风格。目前支持以下风格取值：
        - <auto> 默认
        - <anime> 动画
        - <3d cartoon> 3D卡通
        - <oil painting> 油画
        - <watercolor> 水彩
        - <sketch> 素描
        - <chinese painting> 中国画
        - <flat illustration> 扁平插画
    """
    def __init__(self, model: str=None, **kwargs):
        super().__init__(model=(model or "wanx-x-painting"), **kwargs)

    @property
    def aigc_base_url(self):
        return f"{DASHSCOPE_BASE_URL}/services/aigc/image2image/image-synthesis"

    def get_data(self, input: Dict[str, Any], parameters: Dict[str, Any]=None):
        return {
            "model": self.model,
            "input": {
                "base_image_url": self.confirm_content_url(input, "base_image", "_url"),
                "mask_image_url": self.confirm_content_url(input, "mask_image", "_url"),
                "prompt": input.get("prompt")
            },
            "parameters": parameters
        }

class BackgroundWanx(Text2ImageWanx):
    """
    通义万相-图像背景生成可以基于输入的前景图像素材拓展生成背景信息，实现自然的光影融合效果，与细腻的写实画面生成。

    :input.base_image_url: 透明背景的主体图像URL。
    需要为带透明背景的RGBA 四通道图像，支持png格式，分辨率长边不超过2048像素。
    输出图像的分辨率与该输入图相同。
    """
    def __init__(self, model: str=None, **kwargs):
        super().__init__(model=(model or "wanx-background-generation-v2"), **kwargs)

    @property
    def aigc_base_url(self):
        return f"{DASHSCOPE_BASE_URL}/services/aigc/background-generation/generation/"

    def get_data(self, input: Dict[str, Any], parameters: Dict[str, Any]=None):
        return {
            "model": self.model,
            "input": {
                "base_image_url": self.confirm_content_url(input, "base_image", "_url"),
                "ref_image_url": self.confirm_content_url(input, "ref_image", "_url"),
                "reference_edge": {
                    "foreground_edge": self.confirm_content_url(input, "foreground", "_edge"),
                    "background_edge": self.confirm_content_url(input, "background", "_edge")
                },
                "neg_ref_prompt": input.get("neg_ref_prompt", "")[:70] or None,
                "ref_prompt": input.get("ref_prompt", "")[:70] or None,
                "title": input.get("title", "")[:8] or None,
                "sub_title": input.get("sub_title", "")[:10] or None
            },
            "parameters": parameters
        }


class AnyTextWanx(Text2ImageWanx):
    """
    通义万相-AnyText图文融合，支持图文生成和文字编辑功能，可广泛应用于电商海报、Logo设计、创意涂鸦、表情包、儿童绘本等诸多场景。

    :input.prompt: 半角双引号前需加反斜杠转义字符。
    """
    def __init__(self, model: str=None, **kwargs):
        super().__init__(model=(model or "wanx-anytext-v1"), **kwargs)

    @property
    def aigc_base_url(self):
        return f"{DASHSCOPE_BASE_URL}/services/aigc/anytext/generation/"

    def get_data(self, input: Dict[str, Any], parameters: Dict[str, Any]=None):
        default_mask = "http://public-vigen-video.oss-cn-shanghai.aliyuncs.com/yuxiang.tyx/test_anytext/gen1.png"
        return {
            "model": self.model,
            "input": {
                "mask_image_url": self.confirm_content_url(input, "mask_image", "_url") or default_mask,
                "base_image_url": self.confirm_content_url(input, "base_image", "_url"),
                "reference_edge": {
                    "foreground_edge": self.confirm_content_url(input, "foregrund", "_edge"),
                    "background_edge": self.confirm_content_url(input, "background", "_edge")
                },
                "prompt": input.get("prompt", None),
                "appended_prompt": input.get("appended_prompt", None),
                "negative_prompt": input.get("negative_prompt", None),
                "layout_priority": input.get("layout_priority", "vertical"),
                "text_position_revise": input.get("text_position_revise", False),
                "title": input.get("title", "")[:8] or None,
                "sub_title": input.get("sub_title", "")[:10] or None
            },
            "parameters": parameters
        }


class SketchWanx(Text2ImageWanx):
    """
    通义万相-涂鸦作画通过手绘任意内容加文字描述，即可生成精美的涂鸦绘画作品，
    作品中的内容在参考手绘线条的同时，兼顾创意性和趣味性。
    涂鸦作画支持扁平插画、油画、二次元、3D卡通和水彩5种风格，可用于创意娱乐、辅助设计、儿童教学等场景。

    :parameters.size: 生成图像的分辨率，目前仅支持：
        - 768*768 默认值

    :parameters.style: 输出图像的风格，目前支持以下风格：
        - <3d cartoon> 3D 卡通
        - <anime> 二次元
        - <oil painting> 油画
        - <watercolor> 水彩
        - <flat illustration> 扁平插画
    """
    def __init__(self, model: str=None, **kwargs):
        super().__init__(model=(model or "wanx-sketch-to-image-lite"), **kwargs)

    @property
    def aigc_base_url(self):
        return f"{DASHSCOPE_BASE_URL}/services/aigc/image2image/image-synthesis/"

    def get_data(self, input: Dict[str, Any], parameters: Dict[str, Any]=None):
        return {
            "model": self.model,
            "input": {
                "sketch_image_url": self.confirm_content_url(input, "sketch_image", "_url"),
                "prompt": input.get("prompt", "")[:75],
            },
            "parameters": parameters
        }

class WordArt(Text2ImageWanx):
    """
    WordArt锦书-文字纹理生成可以对输入的文字内容或文字图片进行创意设计，根据提示词内容对文字添加材质和纹理，
    实现立体材质、场景融合、光影特效等效果，生成效果精美、风格多样的艺术字，结合背景可以直接作为文字海报使用。

    input.image 和 input.text 二选一

    :input.image.image_url: 图像要求：
        黑底白字, 图片大小小于5M, 图像格式推荐jpg/png/jpeg/bmp, 长宽比不大于2, 最大边长小于等2048;
        若选择了input.image, 此字段为必须字段

    :input.text.font_name: 当使用input.text时, input.text.ttf_url和input.text.font_name 需要二选一
        - dongfangdakai 阿里妈妈东方大楷, 默认
        - puhuiti_m 阿里巴巴普惠体
        - shuheiti 阿里妈妈数黑体
        - jinbuti 钉钉进步体
        - kuheiti 站酷酷黑体
        - kuaileti 站酷快乐体
        - wenyiti 站酷文艺体
        - logoti 站酷小薇LOGO体
        - cangeryuyangti_m 站酷仓耳渔阳体
        - siyuansongti_b 思源宋体
        - siyuanheiti_m 思源黑体
        - fangzhengkaiti 方正楷体

    :input.text.output_image_ratio: 文字输入的图片的宽高比:
        - 1:1 默认
        - 16:9
        - 9:16

    :parameters.image_short_size: 
    生成的图片短边的长度, 默认为704, 取值范围为[512, 1024],
    若输入数值非64的倍数, 则最终取值为不大于该数值的能被64整除的最大数;
    若输入为图片，输出图片的宽高比和输入图片保持一致.
    """
    def __init__(self, model: str=None, **kwargs):
        super().__init__(model=(model or "wordart-texture"), **kwargs)

    @property
    def aigc_base_url(self):
        return f"{DASHSCOPE_BASE_URL}/services/aigc/wordart/texture"

    def get_data(self, input: Dict[str, Any], parameters: Dict[str, Any]=None):
        image = {
            "image_url": self.confirm_content_url(input, "image", "_url")
        }

        text = {
            "text_content": input.get("text_content", "")[:6] or None,
            "ttf_url": self.confirm_content_url(input, "ttf", "_url"),
            "font_name": input.get("font_name", None),
            "output_image_ratio": input.get("output_image_ratio", '1:1')
        }

        return {
            "model": self.model,
            "input": {
                "image": image if image.get("image_url", None) else None,
                "text": text if text.get("text_content", None) else None,
                "prompt": input.get("prompt")[:200],
                "texture_style": input.get("texture_style", "material")
            },
            "parameters": {
                "n": 1,
                **parameters
            }
        }

class WordArtSemantic(Text2ImageWanx):
    """
    WordArt锦书-文字变形可以对输入的文字边缘轮廓进行创意变形，
    根据提示词内容进行边缘变化，实现一种字体的更多种创意用法，
    返回带有文字内容的黑底白色蒙版图。
    """
    def __init__(self, model: str=None, **kwargs):
        super().__init__(model=(model or "wordart-texture"), **kwargs)

    @property
    def aigc_base_url(self):
        return f"{DASHSCOPE_BASE_URL}/services/aigc/wordart/semantic"

    def get_data(self, input: Dict[str, Any], parameters: Dict[str, Any]=None):
        return {
            "model": self.model,
            "input": {
                "text": input.get("text", None),
                "prompt": input.get("prompt")[:200],
                "texture_style": input.get("texture_style", "material")
            },
            "parameters": {
                "n": 1,
                **parameters
            }
        }



from typing import List, Union, Callable, Generator, AsyncGenerator
from ...utils import minify_text
from ...types import VectorDB, EventBlock, Document
from ...io import log, alog
from ...core.runnable import MarkMeta
from ...utils import raise_invalid_params

import time
import numpy as np

class FaissDB(VectorDB):
    @classmethod
    def allowed_params(cls):
        return {
            "train": "是否在加载数据时训练模型",
            **VectorDB.allowed_params()
        }

    def __init__(self, train: bool=None, **kwargs):
        raise_invalid_params(kwargs, self.__class__.allowed_params())

        super().__init__(**kwargs)
        try:
            import faiss
        except ImportError:
            raise ImportError(
                "Could not import faiss package. "
                "Please install it via 'pip install -U faiss-cpu faiss-gpu'"
            )

        self.train = train if train is not None else True
        self.index = faiss.IndexFlatL2(self.dim) 

        self.documents = self.embeddings.last_output[:]
        self.load_documents(self.documents)
    
    def load_text(self, text: str, source: str=None, **kwargs):
        mm = MarkMeta(**kwargs)
        docs = mm.load_text(text, source)

        self.load_documents(docs)

    def load_documents(self, docs: List[Document], **kwargs):
        vectors = self._process_embeddings(docs)
        if vectors is not None and len(vectors) > 0:
            if self.train:
                self.index.train(vectors)
            self.index.add(vectors)
            self.documents.extend(docs)

    def load(
        self,
        dir: str=None,
        verbose: bool = False,
        handlers: List[Union[Callable, Generator, AsyncGenerator]] = None,
        call_func: Callable = None,
        **kwargs
    ):
        """
        使用 MarkMeta 的 load 方法从指定目录加载文件。
        """
        # 记录文档来源
        self.sources.append(dir)

        mm = MarkMeta(dir=dir, **kwargs)
        mm(verbose=verbose, handlers=handlers, action="load", **kwargs)

        vectors = self._process_embeddings(mm.last_output)
        if vectors is not None and len(vectors) > 0:
            if self.train:
                self.index.train(vectors)
            self.index.add(vectors)
            self.documents.extend(mm.last_output)

    def _process_embeddings(self, docs: List[Document]):
        """
        处理 embeddings.last_output 并返回 NumPy 数组。
        """
        embedded_docs = [d for d in docs if 'embeddings' in d.meta]
        not_embedded_docs = [d for d in docs if 'embeddings' not in d.meta]

        if not_embedded_docs:
            new_embeddings = self.embeddings(not_embedded_docs, verbose=self.verbose)
            embedded_docs.extend(new_embeddings)

        if embedded_docs:
            vectors = [d.meta['embeddings'] for d in embedded_docs]
            return np.array(vectors, dtype='float32')  # 将列表转换为NumPy数组
        return None

    def query(self, text: str, top_k: int=None, **kwargs) -> List[Document]:
        """
        查询向量。

        由于查询过程需要快速返回，因此不按迭代器返回。
        """
        if len(self.documents) == 0:
            return []
        elif not text or len(text.strip()) == 0:
            return []
        else:
            # 对输入字符串做向量编码并查询
            vectors = [self.embeddings.query(text)]
            vectors = np.array(vectors, dtype='float32')  # 将查询向量转换为NumPy数组
            distances, indices = self.index.search(vectors, top_k or self.top_k or 5)

            # 筛选出文档
            valid_indices = indices[0][indices[0] >= 0]
            results = [(distances[0][i], self.documents[valid_indices[i]]) for i in range(len(valid_indices))]

            # 按距离排序
            results.sort(key=lambda x: x[0])
            for distance, doc in results:
                doc.meta["distance"] = distance

            # 返回的结果是 Document 列表
            return [doc for _, doc in results]



from typing import Any
import json
import hashlib
import numpy as np
import pandas as pd
import copy
import uuid
from datetime import datetime

from ..config import get_env, get_ascii_color_code

class EventBlock():
    def __init__(
        self, 
        block_type: str, 
        content: Any,
        content_id: str=None,
        created_at: datetime=None, 
        calling_info: dict=None,
        runnable_info: dict=None
    ):
        self.id = str(uuid.uuid1().hex)
        self.content_id = content_id or str(uuid.uuid1().hex)
        self.content = content
        self.block_type = block_type.lower()
        self.created_at = created_at or datetime.now()
        self.calling_info = calling_info or {}
        self.runnable_info = runnable_info or {}

    def __str__(self):
        return self.text
    
    def __repr__(self):
        return f"EventBlock(block_type=<{self.block_type}>, content=<{self.text}>)"
    
    @property
    def json(self):
        return json.dumps({
            "id": self.id,
            "content_id": self.content_id,
            "block_type": self.block_type,
            "content": self.text,
            "created_at": self.created_at.isoformat(),
            "calling_info": self.calling_info,
            "runnable_info": self.runnable_info,
        }, ensure_ascii=False)

    @property
    def text(self):
        """
        兼容多模态时返回图像、视频等情况
        """
        if isinstance(self.content, str):
            return self.content
        elif isinstance(self.content, list):
            items = []
            for item in self.content:
                if isinstance(item, dict) and "text" in item:
                    items.append(item["text"])
                else:
                    items.append(json.dumps(item, ensure_ascii=False))
            return ",".join(items)
        else:
            return str(self.content)
    
    @property
    def text_with_print_color(self):
        color_mapping = {
            # 过程片段
            'chunk': "ILLUFLY_COLOR_CHUNK",
            'tools_call_chunk': "ILLUFLY_COLOR_CHUNK",
            'tool_resp_chunk': "ILLUFLY_COLOR_CHUNK",
            # 将片段累积后的输出
            'final_text': "ILLUFLY_COLOR_FINAL",
            'final_tools_call': "ILLUFLY_COLOR_FINAL",
            'final_tool_resp': "ILLUFLY_COLOR_FINAL",
            # 直接输出的文本
            'text': "ILLUFLY_COLOR_TEXT",
            'image_url': "ILLUFLY_COLOR_TEXT",
            # 警告信息
            'warn': "ILLUFLY_COLOR_WARN",
            # 其他信息
            'unknown': "ILLUFLY_COLOR_INFO",
        }

        env_var_name = color_mapping.get(self.block_type, "ILLUFLY_COLOR_INFO")
        color = get_env(env_var_name)
        return get_ascii_color_code(color) + self.text + "\033[0m"

class ResponseBlock(EventBlock):
    """
    用于在使用生成器的函数之间传递返回值。
    """
    def __init__(self, resp: Any, *args, **kwargs):
        super().__init__("RESPONSE", *args, **kwargs)

class EndBlock(EventBlock):
    def __init__(self, output_text: str):
        tail_text = self.create_chk_block(output_text)
        super().__init__("END", tail_text)

    def create_chk_block(self, output_text: str):
        """
        生成哈希值
        """
        # 移除前后空格以确保唯一性
        trimmed_output_text = output_text.strip()
        hash_object = hashlib.sha256(trimmed_output_text.encode())
        # 获取十六进制哈希值
        hash_hex = hash_object.hexdigest()
        # 转换为8位数字哈希值
        hash_code = int(hash_hex, 16) % (10 ** 8)

        tail = f'【{get_env("ILLUFLY_AIGC_INFO_DECLARE")}，{get_env("ILLUFLY_AIGC_INFO_CHK")} {hash_code}】'

        return tail

class NewLineBlock(EventBlock):
    def __init__(self, *args, **kwargs):
        super().__init__("new_line", "", *args, **kwargs)




from typing import List
from ......io import EventBlock, NewLineBlock
from ....message import Messages
from .base import BaseToolCalling
import json

class OpenAIToolsCalling(BaseToolCalling):
    """
    OpenAI 风格工具调用。
    """
    def handle(self, final_tools_call:str, short_term_memory:Messages, long_term_memory:Messages):
        # 由于 OpenAI 风格工具回调是从文本之外的参数返回的，因此额外追加一次 EventBlock
        final_tools_call_text = json.dumps(final_tools_call, ensure_ascii=False)
        yield NewLineBlock()
        yield EventBlock("final_tools_call", final_tools_call_text)

        for index, tool in enumerate(final_tools_call):
            tools_call_message = [{
                "role": "assistant",
                "content": "",
                "tool_calls": [tool]
            }]

            if self.tools_to_exec:
                # 记忆追加
                short_term_memory.extend(tools_call_message)
                long_term_memory.extend(tools_call_message)

                for block in self.execute_tool(tool):
                    if isinstance(block, EventBlock) and block.block_type == "final_tool_resp":
                        tool_resp = block.text
                        tool_resp_message = [{
                            "tool_call_id": tool['id'],
                            "role": "tool",
                            "name": tool['function']['name'],
                            "content": tool_resp
                        }]
                        # 短期记忆追加
                        short_term_memory.extend(tool_resp_message)
                        # 长期记忆追加
                        long_term_memory.extend(tool_resp_message)
                    yield block

    async def async_handle(self, final_tools_call: str, short_term_memory: Messages, long_term_memory: Messages):
        final_tools_call_text = json.dumps(final_tools_call, ensure_ascii=False)
        yield EventBlock("final_tools_call", final_tools_call_text)

        for index, tool in enumerate(final_tools_call):
            tools_call_message = [{
                "role": "assistant",
                "content": "",
                "tool_calls": [tool]
            }]
            if self.tools_to_exec:
                # 记忆追加
                short_term_memory.extend(tools_call_message)
                long_term_memory.extend(tools_call_message)

                async for block in self.async_execute_tool(tool):
                    if isinstance(block, EventBlock) and block.block_type == "final_tool_resp":
                        tool_resp = block.text
                        tool_resp_message = [{
                            "tool_call_id": tool['id'],
                            "role": "tool",
                            "name": tool['function']['name'],
                            "content": tool_resp
                        }]
                        # 短期记忆追加
                        short_term_memory.extend(tool_resp_message)
                        # 长期记忆追加
                        long_term_memory.extend(tool_resp_message)
                    yield block    


from typing import List, Union, Optional
from .....config import get_env
from .....utils import extract_text, minify_text
from .....io import EventBlock
from ....document import Document
from ...base import Runnable
from ...vectordb import VectorDB
from ...prompt_template import PromptTemplate
from ...markmeta import MarkMeta
from ..base import BaseAgent

class Retriever(BaseAgent):
    """
    检索器。

    实现流程包括：意图理解 -> 数据查询 -> 整理结果

    意图理解主要是根据问题索引向量库，或对话模型，将原始问题发散，生成更多提问角度；
    数据查询主要通过 关系数据库、向量数据库、图数据库等实现。

    控制参数说明：
    - 决定广度 question_count: int=3,
    - 决定深度 search_count: int=20,
    - 决定句法关联度 rerank_count: int=5,
    """
    def __init__(
        self,
        translators: List[Runnable]=None,
        searchers: List[Runnable]=None,
        reranker: Runnable=None,
        question_count: int=3,
        search_count: int=20,
        rerank_count: int=5,
        **kwargs
    ):
        if translators and not isinstance(translators, list):
            translators = [translators]
        if searchers and not isinstance(searchers, list):
            searchers = [searchers]

        self.translators = translators or []
        self.searchers = searchers or []
        self.reranker = reranker

        self.question_count = question_count
        self.search_count = search_count
        self.rerank_count = rerank_count

        super().__init__(**kwargs)

        for item in self.translators:
            from ..chat import ChatAgent
            if isinstance(item, ChatAgent) and not item.init_memory:
                item.reset_init_memory(
                    PromptTemplate(
                        "RAG/Q_GEN",
                        binding_map={"count": self.question_count}
                    )
                )
            elif isinstance(item, VectorDB):
                if not item.top_k:
                    item.top_k = question_count
                if not item.documents:
                    item.load(dir=get_env("ILLUFLY_IDENT"))

        for item in self.searchers:
            if isinstance(item, VectorDB):
                if not item.top_k:
                    item.top_k = search_count
                if not item.documents:
                    item.load(dir=get_env("ILLUFLY_DOCS"))

    def call(self, query: str, **kwargs) -> List[dict]:
        """
        检索过程的具体实现包括如下过程：
        1. 意图理解：如果提供 translators，则分别转换问题，并合并为问题列表
        2. 初步查询：如果提供 searchers，则使用 searchers 分别查询问题列表
        3. 整理结果：如果提供 reranker，则使用 reranker 重新排序结果
        """
        understood_queries = set()
        understood_queries.add(query)
        understood_queries.update(self.translate(query, **kwargs))
        yield from self.search(understood_queries, **kwargs)

    def translate(self, query: str, **kwargs) -> List[str]:
        """
        意图理解，将问题翻译为进一步检索的问题列表。
        这个过程可以是：拆解问题、发散问题、回溯问题、模拟回答等，具体取决于 translator 的类型和执行情况。
        translator 可以是 ChatAgent, VectorDB 或 BaseAgent,
        而返回结果可以是 字符串列表、JSON 文本、MarkMeta 文本或 Document 列表等几种格式。
        """
        from ..chat.base import ChatAgent

        translated_queries = set()
        for translator in self.translators:
            if isinstance(translator, ChatAgent):
                mm = MarkMeta()
                output_text = translator(query, new_chat=True)
                valid_text = extract_text(output_text, ("```markdown", "```"))
                docs = mm.load_text(valid_text)
                # mm.save()
                for doc in docs:
                    translated_queries.add(doc.text)

            elif isinstance(translator, VectorDB):
                top_k = self.question_count
                translated_results = translator(query)
                for result in translated_results:
                    if isinstance(result, str):
                        translated_queries.add(result)
                    elif isinstance(result, Document):
                        translated_queries.add(result.text)
                    else:
                        raise ValueError("Unknown translation result type")

            else:
                raise ValueError("Unknown translator type")

        return translated_queries

    def search(self, queries: List[str], **kwargs) -> List[dict]:
        """
        初步查询。
        """
        results = set()
        top_k = self.search_count
        for query in queries:
            for searcher in self.searchers:
                yield EventBlock("agent", f"由 {searcher.name} 检索问题：{query}")

                search_results = searcher(query, top_k=top_k, **kwargs)
                for result in search_results:
                    if isinstance(result, str):
                        yield EventBlock("info", f"检索结果[-]：{result}")
                    elif isinstance(result, Document):
                        yield EventBlock("info", f"检索结果[{result.meta['distance']}]：{minify_text(result.text, 30)}")
                    else:
                        raise ValueError("Unknown search result type")

                if search_results:
                    if self.reranker:
                        yield EventBlock("agent", f"由 {self.reranker.name} 重新排序检索结果")
                        rerank_results = self.reranker(query, search_results, top_k=self.rerank_count, **kwargs)
                        for doc in rerank_results:
                            if isinstance(doc, Document):
                                yield EventBlock("info", f"重新排序结果[{doc.meta['rerank_score']}]：{minify_text(doc.text)}")
                            else:
                                raise ValueError("Unknown rerank result type")
                    else:
                        rerank_results = search_results[:self.rerank_count]

                    for doc in rerank_results:
                        if isinstance(doc, Document):
                            results.add(doc)
                        elif isinstance(doc, str):
                            results.add(Document(doc))
                        else:
                            raise ValueError("Unknown search result type")
        self._last_output = results



"""
    提示语模板采用了`mustache`语法，这是一种逻辑无关的模板语法，包括以下特性：
    1. **变量替换**：使用`{{variable}}`来插入变量的值。
    2. **注释**：使用`{{! comment }}`来添加不会在渲染结果中显示的注释。
    3. **条件判断**：使用`{{#condition}}{{/condition}}`来进行条件判断，如果条件为真，则渲染内部的内容。
    4. **部分模板**：使用`{{> partial}}`来包含另一个模板文件。
    ... 还有其他一些可用的特性
"""

from typing import List, Dict, Any, Union, Set
from importlib.resources import read_text, is_resource, contents
from chevron.renderer import render as mustache_render
from chevron.tokenizer import tokenize as mustache_tokenize
from ..config import get_env

import os
import re
import json
from pathlib import Path

PROMPT_WRITING_BASE = 'illufly.__PROMPT_TEMPLATES__'

def find_resource_template(*seg_path):
    """
    过滤出提示语模板所在的目录清单。
    """
    package_path = ".".join([PROMPT_WRITING_BASE, *seg_path])
    all_resources = contents(package_path)
    return [r for r in all_resources if not is_resource(PROMPT_WRITING_BASE, r)]

def load_resource_template(template_id: str):
    """
    从python包资源文件夹加载提示语模板。
    """
    normalized_template_id = template_id.replace(os.sep, '.').replace('/', '.')
    parts = normalized_template_id.split('.')
    if parts[-1] not in find_resource_template(*parts[:-1]):
        raise ValueError(f"<{template_id}> template_id 不存在！")

    def _get_template_str(res_file: str):
        resource_folder = f'{PROMPT_WRITING_BASE}.{normalized_template_id}'
        if is_resource(resource_folder, res_file):
            return read_text(resource_folder, res_file)
        resource_base = PROMPT_WRITING_BASE
        if is_resource(resource_base, res_file):
            return read_text(resource_base, res_file)
        return ''

    prompt_str = _get_template_str('main.mu')

    # 替换 {{>include_name}} 变量
    include_dict = {}
    matches = re.findall(r'{{>(.*?)}}', prompt_str)
    for part_name in matches:
        included_str = _get_template_str(f'{part_name.strip()}.mu')
        if included_str:
            include_dict[part_name] = included_str
        else:
            raise RuntimeError(f"无法在 {template_id} 中找到 {part_name}.mu！")
    for part_name, part_str in include_dict.items():
        prompt_str = prompt_str.replace(f"{{>{part_name}}}", part_str)

    return prompt_str

def _find_prompt_file(template_id: str, file_name: str, template_folder: str = None, sep: str = None, all_path: List[str] = None, force: bool = False):
    sep = sep or os.sep
    if all_path is None:
        all_path = []
    normalized_template_id = template_id.replace('/', sep).strip(f'{sep} ').replace('\\', sep)

    prompt_folder = Path(template_folder or "") / Path(normalized_template_id)
    prompt_folder = prompt_folder.as_posix() if sep != os.sep else str(prompt_folder)

    for ext in ['mu', 'mustache', 'txt']:
        file_path = prompt_folder / f'{file_name}.{ext}' if isinstance(prompt_folder, Path) else os.path.join(prompt_folder, f'{file_name}.{ext}')
        if os.path.exists(file_path):
            return file_path

    all_path.append(prompt_folder)
    if not normalized_template_id:
        if force:
            raise ValueError(f"无法在 [ {', '.join(all_path)} ] 中找到 {file_name}(.mu, .mustache, .txt) 文件！")
        else:
            return None
    parent_id = Path(normalized_template_id).parent
    return _find_prompt_file(parent_id.as_posix() if sep != os.sep else parent_id.name, file_name, template_folder, sep, all_path, force)

def load_prompt_template(template_id: str, template_folder: str = None):
    """
    从模板文件夹加载提示语模板。

    提示语模板中的约定：
    1. 加载模板时，从模板路径开始寻找，如果找不到会向上查找文件，直至模板根文件夹
    2. 提示语模板的主文件是 main.mu
    3. 预处理 main.mu 中的 {{>include_name}} 语法
    4. 如果找不到 main.mu 文件，则从资源文件夹中加载
    """
    main_prompt = _find_prompt_file(template_id, 'main', template_folder)

    if main_prompt:
        with open(main_prompt, 'r', encoding='utf-8') as f:
            prompt_str = f.read()

            # 替换 {{>include_name}} 变量
            include_dict = {}
            matches = re.findall(r'{{>(.*?)}}', prompt_str)
            for part_name in matches:
                part_file = _find_prompt_file(template_id, part_name.strip(), template_folder)
                if part_file:
                    with open(part_file, 'r', encoding='utf-8') as pf:
                        include_dict[part_name] = pf.read()
                else:
                    raise RuntimeError(f"无法找到包含文件 {part_name}.mu")
            for part_name, part_str in include_dict.items():
                prompt_str = prompt_str.replace(f"{{>{part_name}}}", part_str)

            return prompt_str
    else:
        prompt_str = load_resource_template(template_id)
        if prompt_str:
            return prompt_str

    raise ValueError(f'无法构建模板：{template_id}')

def get_template_variables(template_text: str):
    vars: Set[str] = set()
    section_depth = 0
    for token_type, key in mustache_tokenize(template_text):
        if token_type == "end":
            section_depth -= 1
        elif (
            token_type in ("variable", "section", "inverted section", "no escape")
            and key != "."
            and section_depth == 0
        ):
            vars.add(key.split(".")[0])
        if token_type in ("section", "inverted section"):
            section_depth += 1
    return vars

def clone_prompt_template(template_id: str, template_folder: str = None):
    """
    克隆提示语模板。
    根据指定 template_id 将文件夹和文件拷贝到本地 template_folder 位置。
    
    如果已经存在，就不再覆盖已修改的模板成果。
    """
    local_folder = Path(template_folder or get_env("ILLUFLY_PROMPT_TEMPLATE_LOCAL_FOLDER"))
    template_path = local_folder / Path(template_id.replace('.', os.sep))
    template_path_str = str(template_path)

    # 如果已经存在，就不再覆盖已修改的模板成果。
    if template_path.exists():
        if any(template_path.iterdir()):
            raise ValueError(f"模板文件夹 [{template_path_str}] 非空！")
    else:
        template_path.mkdir(parents=True, exist_ok=True)

    normalized_template_id = template_id.replace(os.sep, '.')
    parts = normalized_template_id.split('.')
    if parts[-1] not in find_resource_template(*parts[:-1]):
        raise ValueError(f"<{template_id}> template_id 不存在！")

    def _get_template_str(res_file: str):
        resource_folder = f'{PROMPT_WRITING_BASE}.{normalized_template_id}'
        if is_resource(resource_folder, res_file):
            return read_text(resource_folder, res_file)
        resource_base = PROMPT_WRITING_BASE
        if is_resource(resource_base, res_file):
            return read_text(resource_base, res_file)
        return ''

    prompt_str = _get_template_str('main.mu')
    if not prompt_str:
        raise ValueError(f'main.mu 必须存在于 {template_id} 中！')

    main_mu_path = template_path / 'main.mu'
    with open(main_mu_path, 'w', encoding='utf-8') as f:
        f.write(prompt_str)

        # 保存替换 {{>include_name}} 变量文件
        matches = re.findall(r'{{>(.*?)}}', prompt_str)
        for part_name in matches:
            part_str = _get_template_str(f'{part_name.strip()}.mu')
            if part_str:
                part_mu_path = template_path / f'{part_name}.mu'
                with open(part_mu_path, 'w', encoding='utf-8') as pf:
                    pf.write(part_str)
            else:
                raise RuntimeError(f"无法在 {template_id} 中找到 {part_name}.mu")

    return template_path_str



import os
import platform

def get_env(key: str=None):
    """
    获取环境变量，如果没有设置就根据默认值清单返回默认值。
    """
    # 根据操作系统设置固定的临时目录
    if platform.system() == "Windows":
        FIXED_TEMP_DIR = os.path.join(os.getenv('TEMP'), "__ILLUFLY__")
    elif platform.system() == "Darwin":  # macOS
        FIXED_TEMP_DIR = os.path.join(os.getenv('TMPDIR', '/tmp'), "__ILLUFLY__")
    else:  # Linux and other Unix-like systems
        FIXED_TEMP_DIR = os.path.join('/tmp', "__ILLUFLY__")

    default_values = {
        # 文件夹配置
        "ILLUFLY_DOCS": "__DOCS__", # 从这个目录自动加载 RAG 资料，为避免误将不必要的资料做 RAG，因此需要设置为 __DOCS__
        "ILLUFLY_RESOURCE": "./", # 根据这个媒体文件夹上传文件，因此需要设置为当前目录
        "ILLUFLY_LOCAL_FILE_MEMORY": os.path.join(FIXED_TEMP_DIR, "LOCAL_FILE_MEMORY"), # LocalFileMemory 记忆存储目录
        "ILLUFLY_CHAT_LEARN": os.path.join(FIXED_TEMP_DIR, "CHART_LEARN"), # ChatLearn 自我进化经验的存储目录
        "ILLUFLY_IDENT": os.path.join(FIXED_TEMP_DIR, "IDENT"), # 问题转换、意图识别的经验存储目录
        "ILLUFLY_CACHE_EMBEDDINGS": os.path.join(FIXED_TEMP_DIR, "CACHE_EMBEDDINGS"), # 缓存 Embeddings 存储目录
        "ILLUFLY_UPLOAD_CACHE": os.path.join(FIXED_TEMP_DIR, "UPLOAD_CACHE"), # 上传缓存目录

        # 提示语
        "ILLUFLY_PROMPT_TEMPLATE_LOCAL_FOLDER": "__PROMPTS__",
        "ILLUFLY_FINAL_ANSWER_PROMPT": "最终答案",
        "ILLUFLY_FINAL_ANSWER_START": "<final_answer>",
        "ILLUFLY_FINAL_ANSWER_END": "</final_answer>",
        "ILLUFLY_CONTEXT_START": "<context>",
        "ILLUFLY_CONTEXT_END": "</context>",
        "ILLUFLY_KNOWLEDGE_START": "<knowledge>",
        "ILLUFLY_KNOWLEDGE_END": "</knowledge>",
        "ILLUFLY_OUTPUT_START": "```markdown",
        "ILLUFLY_OUTPUT_END": "```",

        # 结束语
        "ILLUFLY_AIGC_INFO_DECLARE": "内容由AI生成，其观点仅代表创作者个人立场",
        "ILLUFLY_AIGC_INFO_CHK": "可联系服务商查验校验码",

        # 扩写标签
        "ILLUFLY_OUTLINE_START": "<OUTLINE>",
        "ILLUFLY_OUTLINE_END": "</OUTLINE>",

        # log 颜色
        "ILLUFLY_COLOR_DEFAULT": "黑色",
        "ILLUFLY_COLOR_INFO": "蓝色",
        "ILLUFLY_COLOR_TEXT": "黄色",
        "ILLUFLY_COLOR_WARN": "红色",
        "ILLUFLY_COLOR_CHUNK": "绿色",
        "ILLUFLY_COLOR_FINAL": "青色",
        "ILLUFLY_COLOR_FRONT_MATTER": "品红",

        # HTTP
        "HTTP_CHECK_RESULT_SECONDS": 2,

        # DashScope
        "DASHSCOPE_BASE_URL": "https://dashscope.aliyuncs.com/api/v1",
        # Zhipu
        "ZHIPUAI_API_TOOLS_BASE_URL": "https://open.bigmodel.cn/api/paas/v4/tools"
    }
    if key:
        if key not in default_values:
            raise ValueError(f"Environ Value [{key}] Not Exist !!!")
        else:
            return os.getenv(key) or default_values[key]
    else:
        return default_values

def get_ascii_color_code(color_name: str):
    """
    获取 ASCII 颜色代码。
    """
    colors = {
        "黑色": "\033[30m",
        "红色": "\033[31m",
        "绿色": "\033[32m",
        "黄色": "\033[33m",
        "蓝色": "\033[34m",
        "品红": "\033[35m",
        "青色": "\033[36m",
        "白色": "\033[37m",
        "重置": "\033[0m" 
    }
    return colors.get(color_name, '黑色')



from typing import Iterable, Dict, Any, List, Union, Tuple
import re
import time
import random
import yaml
import copy
from mistune import markdown
from mistune.renderers.markdown import MarkdownRenderer
from mistune.core import BlockState
from .....document import Document
from ......config import get_env

class SegmentsRenderer(MarkdownRenderer):
    def __init__(self, doc_id_generator, source: str=None):
        super().__init__()
        self.doc_id_generator = doc_id_generator
        self.source = source or 'unknown'

    def __call__(self, tokens: Iterable[Dict[str, Any]], state: BlockState) -> str:
        documents = []
        for tok in tokens:
            md = self.render_token(tok, state)
            doc_id = next(self.doc_id_generator)
            tok.update({"id": doc_id, "source": self.source})
            if tok['type'] == 'blank_line':
                md = '\n'
            documents.append(Document(text=md, meta=tok))
        return documents

def create_front_matter(dict_data: Dict[str, Any]):
    """
    构造 YAML Front Matter
    """
    if isinstance(dict_data, dict):
        meta = copy.deepcopy(dict_data)
        for e_tag in ['id', 'type']:
            if e_tag in meta:
                meta.pop(e_tag, None)
        for e_tag in ['verbose', 'is_fake']:
            tags = meta.get('args', {})
            if e_tag in tags:
                tags.pop(e_tag, None)
        yaml_str = yaml.safe_dump(meta, allow_unicode=True, sort_keys=False)
        return "---\n" + yaml_str.replace("\n\n", "\n") + "---\n\n"
    else:
        return ''

def fetch_front_matter(text: str):
    """
    提取 YAML Front Matter
    """
    yaml_pattern = re.compile(r'\n*---+\n(.*?)\n---+\n', re.DOTALL)
    yaml_match = yaml_pattern.match(text)
    if yaml_match:
        yaml_front_matter = yaml_match.group(1)
        meta = yaml.safe_load(yaml_front_matter)
        return meta, text[yaml_match.end():]
    else:
        return {}, text

def parse_markdown(text: str, start_tag: str=None, end_tag: str=None, source: str=None):
    """
    你可以修改 start_tag/end_tag, 使用 <<<< ... >>>> 或 {{ ... }} 等其他方案来标记提纲内容，
    默认为 <OUTLINE> ... </OUTLINE> 的形式。
    """
    start_tag = start_tag or get_env("ILLUFLY_OUTLINE_START")
    end_tag = end_tag or get_env("ILLUFLY_OUTLINE_END")
    doc_id_generator = create_document_id()
    pattern = re.compile(r'(.*?)(%s.*?%s)(.*)' % (re.escape(start_tag), re.escape(end_tag)), re.DOTALL)
    documents = []

    # 提取 YAML Front Matter
    meta, text = fetch_front_matter(text)
    if meta:
        doc_id = next(doc_id_generator)
        meta.update({"id": doc_id, "type": "front_matter", "source": source})
        doc = Document(text='', meta=meta)
        documents.append(doc)

    # 从文本中提取标记内的内容
    while start_tag in text and end_tag in text:
        match = pattern.match(text)
        if match:
            before, outline_content, after = match.groups()
            if before:
                documents.extend(markdown(before, renderer=SegmentsRenderer(doc_id_generator, source)))
            doc_id = next(doc_id_generator)
            doc = Document(text=outline_content+"\n\n", meta={"id": doc_id, "type": 'OUTLINE', "source": source})
            documents.append(doc)
            text = after
    if text:
        documents.extend(markdown(text, renderer=SegmentsRenderer(doc_id_generator, source)))
    return documents

def create_document_id():
    counter = 0
    while True:
        timestamp = str(int(time.time()))[-4:]
        random_number = f'{random.randint(0, 999):03}'
        counter_str = f'{counter:03}'
        yield f'{timestamp}-{random_number}-{counter_str}'
        counter = 0 if counter == 999 else counter + 1

def list_markdown(documents: Union[List[Document], List[Tuple[Document, int]]]):
    if len(documents) > 0:
        if isinstance(documents[0], Document):
            docs = documents
        elif isinstance(documents[0], tuple):
            docs = [d for d, index in documents]
        else:
            docs = []
    return [(d.meta['type'][:2] + "-" + d.meta['id'][-7:], d.text) for d in docs if d]



from .base import Runnable
from .selector import Selector, End
from .agent import BaseAgent, ChatAgent, ChatPool, Retriever
from .agent.flow import CoT, ReAct, ReWOO, PlanAndSolve
from .agent.flow import ChatLearn
from .agent.writer import FromOutline
from .agent.data import PandasAgent, MatplotAgent, PythonAgent
from .agent.chat.tools_calling import BaseToolCalling, ToolCall, SubTask, Plans
from .prompt_template import PromptTemplate
from .importer import Importer
from .embeddings import BaseEmbeddings
from .vectordb import VectorDB
from .markmeta import MarkMeta
from .reranker import BaseReranker
from .team import Team


import json
import time
import asyncio

from typing import Callable, Iterable, Union, AsyncIterable
from ..block import EventBlock, NewLineBlock

__CHUNK_BLOCK_TYPES__ = ["text", "chunk", "tool_resp_text", "tool_resp_chunk"]
__NOT_PRINT_BLOCK_TYPES__ = ["final_text", "response"]
__ALLWAYS_PRINT_BLOCK_TYPES__ = ["final_tools_call", "agent", "faq", "warn", "error", "image_url"]

def process_block(block, verbose:bool):
    if isinstance(block, EventBlock):  
        if block.block_type == "new_line":
            print()
        elif block.block_type in __CHUNK_BLOCK_TYPES__:
            print(block.text_with_print_color, end="")
        else:
            if (verbose or block.block_type in __ALLWAYS_PRINT_BLOCK_TYPES__):
                print(f'[{block.block_type.upper()}] {block.text_with_print_color}')
    elif isinstance(block, str):
        print(block)
    else:
        raise ValueError(f"Unknown block type: {block}")

def log(block, verbose: bool=False, **kwargs):
    """
    打印流式日志。
    """
    if isinstance(block, EventBlock) and block.block_type in __NOT_PRINT_BLOCK_TYPES__:
        return
    process_block(block, verbose=verbose)

async def alog(block, verbose: bool=False, **kwargs):
    """
    打印流式日志。
    """
    if isinstance(block, EventBlock) and block.block_type in __NOT_PRINT_BLOCK_TYPES__:
        return
    process_block(block, verbose=verbose)
    await asyncio.sleep(0)




from ..community.openai import ChatOpenAI
from ..community.zhipu import ChatZhipu, ChatZhipuVL
from ..community.dashscope import ChatQwen, ChatQwenVL
from .fake import FakeLLM

__all__ = [
    "ChatOpenAI",
    "ChatZhipu",
    "ChatZhipuVL",
    "ChatQwen",
    "ChatQwenVL",
    "FakeLLM",
]



from typing import List

import os
from ...types import BaseEmbeddings
from ...utils import raise_invalid_params

class HuggingFaceEmbeddings(BaseEmbeddings):
    """使用开源的 SentenceTransformer 模型进行文本向量化"""

    @classmethod
    def allowed_params(cls):
        return {
            "model": "模型名称",
            **BaseEmbeddings.allowed_params()
        }

    def __init__(self, model: str=None, **kwargs):
        raise_invalid_params(kwargs, self.__class__.allowed_params())

        try:
            from sentence_transformers import SentenceTransformer
            import torch
        except ImportError:
            raise ImportError(
                "Could not import sentence-transformers and torch package. "
                "Please install it via 'pip install -U sentence-transformers torch'"
            )

        super().__init__(
            model=model or "all-MiniLM-L6-v2",
            **kwargs
        )
        self.client = SentenceTransformer(self.model, **kwargs)

        # 检查是否有可用的 CUDA GPU
        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.client = self.client.to(self.device)

    def query(self, text: str) -> List[float]:
        """
        查询文本向量。
        """
        return self.embed_documents([text])[0]

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """
        编码文本向量。
        """
        return self.client.encode(texts, convert_to_tensor=False, device=self.device).tolist()



import asyncio
import inspect
import uuid

from typing import Union, List, Dict, Any, Callable, Generator, AsyncGenerator
from abc import ABC, abstractmethod
from functools import partial

from .executor_manager import ExecutorManager
from .binding_manager import BindingManager
from ...io import log, EventBlock, event_stream
from ...utils import filter_kwargs, raise_invalid_params


class Runnable(ABC, ExecutorManager, BindingManager):
    """
    实现基本可运行类，定义了可运行的基本接口。
    只要继承该类，就可以作为智能体的工具使用。

    实现机制：
    - 支持 EventBlock 流式输出句法
    - 实现 __call__ 方法，来简化流输出调用
    - 支持 call 同步方法调用
    - 支持 async_call 方法调用，并在 Runnable 中已实现默认版本
    - 支持 绑定机制

    什么时候直接从 Runnable 继承？
    - 如果需要支持流式输出
    - 仅仅做数据处理，而没有过多服务端调度逻辑

    什么时候转而使用BaseAgent？
    - 如果需要在多智能体中协同处理
    - 如果包含服务端调度逻辑

    calling_id 是 Runnable 的属性，用于标识一次调用。
    如果在 Runnable 中调用另一个 Runnable，则需要将该 Runnable 的 calling_id 传递给被调用的 Runnable。
    这样，即使在一次调用中嵌套了多层的 Runnable，也可以将当次调用中的事件收集到当前调用的上下文，在下次恢复时，
    可以看到当时的完整调用场景。
    """
    @classmethod
    def allowed_params(cls):
        """
        返回当前可用的参数列表。
        """
        return {
            "name": "Runnable 名称，默认为 {类名}.{id}",
            "handlers": "EventBlock 迭代器处理函数列表，默认为 [log]，当调用 call 方法时，会使用该列表中的函数逐个处理 EventBlock",
            "block_processor": "在 yield 之前将 EventBlock 事件转换为新的格式，在 __call__ 方法的输出生成器时使用",
            **ExecutorManager.allowed_params(),
            **BindingManager.allowed_params(),
        }
    
    @classmethod
    def help(cls):
        """
        返回当前可用的参数列表。
        """
        return f'{cls.__doc__}\n\n{cls.__name__} 参数列表：\n' + "\n".join([f"- {k}: {v}" for k, v in cls.allowed_params().items()])

    def __init__(
        self,
        *,
        name: str = None,
        handlers: List[Union[Callable, Generator, AsyncGenerator]] = None,
        block_processor: Callable = None,
        **kwargs
    ):
        """
        Runnable 的构造函数，主要包括：
        - 初始化线程组
        - 工具：作为工具的Runnable列表，在发现工具后是否执行工具的标记等
        """
        raise_invalid_params(kwargs, self.__class__.allowed_params())
        ExecutorManager.__init__(self, **filter_kwargs(kwargs, ExecutorManager.allowed_params()))

        self.name = name or f'{self.__class__.__name__}.{id(self)}'
        self.continue_running = True
        self.handlers = [log] if handlers is None else handlers
        self.block_processor = block_processor
        self.verbose = False

        self.calling_id = None
        self._last_output = None

        BindingManager.__init__(self, **filter_kwargs(kwargs, BindingManager.allowed_params()))

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.name == other.name
        return False

    def __repr__(self):
        if self.__class__.__name__ in self.name:
            return f"<{self.name}>"
        else:
            return f"<{self.__class__.__name__} {self.name}>"

    @property
    def selected(self):
        """
        让 Runnable 对象与 Selector 的访问方式兼容。
        """
        return self

    @property
    def last_output(self):
        return self._last_output

    @property
    def provider_dict(self):
        local_dict = {
            "last_output": self.last_output,
        }
        return {
            **super().provider_dict,
            **{k:v for k,v in local_dict.items() if v is not None},
        }

    def build_calling_id(self):
        return str(uuid.uuid1())

    def __call__(
        self,
        *args,
        verbose: bool = False,
        handlers: List[Union[Callable, Generator, AsyncGenerator]] = None,
        block_processor: Callable = None,
        generator: str = None,
        action: str = None,
        calling_id: str = None,
        **kwargs
    ):
        if action and not hasattr(self, action):
            raise AttributeError(f"方法 '{action}' 不存在于实例中")

        self._last_output = None
        self.calling_id = calling_id or self.build_calling_id()

        block_processor = block_processor or self.block_processor or event_stream
        self.continue_running = True
        _handlers = handlers if handlers is not None else self.handlers
        _verbose = self.verbose or verbose

        if generator == "async" or any(inspect.iscoroutinefunction(handler) for handler in _handlers):
            method = getattr(self, action) if action else self.async_call

            if generator:
                return self.generate_async_call(
                    *args,
                    verbose=_verbose,
                    handlers=_handlers,
                    action_method=method,
                    block_processor=block_processor,
                    **kwargs
                )
            else:
                return self.handle_async_call(
                    *args,
                    verbose=_verbose,
                    handlers=_handlers,
                    action_method=method,
                    **kwargs
                )
        else:
            method = getattr(self, action) if action else self.call

            if generator:
                return self.generate_sync_call(
                    *args,
                    verbose=_verbose,
                    handlers=_handlers,
                    action_method=method,
                    block_processor=block_processor,
                    **kwargs
                )
            else:
                return self.handle_sync_call(
                    *args,
                    verbose=_verbose,
                    handlers=_handlers,
                    action_method=method,
                    **kwargs
                )

    def handle_block(self, block, handlers, verbose, **kwargs):
        if isinstance(block, str):
            block = EventBlock("text", block)
        elif not isinstance(block, EventBlock):
            block = EventBlock("text", str(block))
        block.runnable_info = self.runnable_info

        for handler in handlers:
            if not inspect.iscoroutinefunction(handler):
                handler(block, verbose=verbose, **kwargs)

    async def async_handle_block(self, block, handlers, verbose, **kwargs):
        if isinstance(block, str):
            block = EventBlock("text", block)
        elif not isinstance(block, EventBlock):
            block = EventBlock("text", str(block))
        block.runnable_info = self.runnable_info

        tasks = []
        for handler in handlers:
            resp = handler(block, verbose=verbose, **kwargs)
            if inspect.isawaitable(resp):
                tasks.append(asyncio.create_task(resp))
        if tasks:
            await asyncio.gather(*tasks)

    def generate_sync_call(
        self,
        *args,
        verbose: bool = False,
        calling_id: str = None,
        handlers: List[Union[Callable, Generator, AsyncGenerator]] = None,
        action_method: Callable = None,
        block_processor: Callable = None,
        **kwargs
    ):
        block = EventBlock("runnable", self.name)
        block.runnable_info = self.runnable_info
        self.handle_block(block, handlers, verbose, **kwargs)
        yield block_processor(block, verbose=verbose, **kwargs)

        if isinstance(handlers, list) and all(callable(handler) for handler in handlers):
            resp = action_method(*args, **kwargs)
            if isinstance(resp, Generator):
                for block in resp:
                    if not self.continue_running:
                        return

                    self.handle_block(block, handlers, verbose, **kwargs)
                    block_text = block_processor(block, verbose=verbose, **kwargs)
                    if block_text:
                        yield block_text
            else:
                block = EventBlock("text", str(resp))
                block.runnable_info = self.runnable_info
                self.handle_block(block, handlers, verbose, **kwargs)
                block_text = block_processor(block, verbose=verbose, **kwargs)
                if block_text:
                    yield block_text
        else:
            raise ValueError("handlers 必须是Callable列表")

    def handle_sync_call(
        self,
        *args,
        verbose: bool = False,
        calling_id: str = None,
        handlers: List[Union[Callable, Generator, AsyncGenerator]] = None,
        action_method: Callable = None,
        **kwargs
    ):
        block = EventBlock("runnable", self.name)
        block.runnable_info = self.runnable_info
        self.handle_block(block, handlers, verbose, **kwargs)

        if isinstance(handlers, list) and all(callable(handler) for handler in handlers):
            resp = action_method(*args, **kwargs)
            if isinstance(resp, Generator):
                for block in resp:
                    if not self.continue_running:
                        return
                    self.handle_block(block, handlers, verbose, **kwargs)
            else:
                block = EventBlock("text", str(resp))
                block.runnable_info = self.runnable_info
                self.handle_block(block, handlers, verbose, **kwargs)
        else:
            raise ValueError("handlers 必须是Callable列表")

        return self.last_output

    async def generate_async_call(
        self,
        *args,
        verbose: bool = False,
        calling_id: str = None,
        handlers: List[Union[Callable, Generator, AsyncGenerator]] = None,
        action_method: Callable = None,
        block_processor: Callable = None,
        **kwargs
    ):
        block = EventBlock("runnable", self.name)
        block.runnable_info = self.runnable_info
        await self.async_handle_block(block, handlers, verbose, **kwargs)
        yield block_processor(block, verbose=verbose, **kwargs)

        if isinstance(handlers, list) and all(callable(handler) for handler in handlers):
            resp = action_method(*args, **kwargs)

            if isinstance(resp, AsyncGenerator):
                async for block in resp:
                    if not self.continue_running:
                        return
                    await self.async_handle_block(block, handlers, verbose, **kwargs)
                    block_text = block_processor(block, verbose=verbose, **kwargs)
                    if block_text:
                        yield block_text
                        await asyncio.sleep(0)
            elif isinstance(resp, Generator):
                for block in resp:
                    if not self.continue_running:
                        return
                    await self.async_handle_block(block, handlers, verbose, **kwargs)
                    block_text = block_processor(block, verbose=verbose, **kwargs)
                    if block_text:
                        yield block_text
                        await asyncio.sleep(0)
            else:
                block = EventBlock("text", str(resp))
                block.runnable_info = self.runnable_info
                await self.async_handle_block(block, handlers, verbose, **kwargs)
                block_text = block_processor(block, verbose=verbose, **kwargs)
                if block_text:
                    yield block_text
                    await asyncio.sleep(0)
        else:
            raise ValueError("handlers 必须是Callable列表")

    async def handle_async_call(
        self,
        *args,
        verbose: bool = False,
        calling_id: str = None,
        handlers: List[Union[Callable, Generator, AsyncGenerator]] = None,
        action_method: Callable = None,
        **kwargs
    ):
        block = EventBlock("runnable", self.name)
        block.runnable_info = self.runnable_info
        await self.async_handle_block(block, handlers, verbose, **kwargs)

        if isinstance(handlers, list) and all(callable(handler) for handler in handlers):
            resp = action_method(*args, **kwargs)
            if isinstance(resp, AsyncGenerator):
                async for block in resp:
                    if not self.continue_running:
                        return
                    await self.async_handle_block(block, handlers, verbose, **kwargs)
            elif isinstance(resp, Generator):
                for block in resp:
                    if not self.continue_running:
                        return
                    await self.async_handle_block(block, handlers, verbose, **kwargs)
            else:
                block = EventBlock("text", str(resp))
                block.runnable_info = self.runnable_info
                await self.async_handle_block(block, handlers, verbose, **kwargs)
        else:
            raise ValueError("handlers 必须是Callable列表")

        return self.last_output

    @property
    def is_running(self):
        return self.continue_running
    
    @property
    def runnable_info(self):
        return {
            "name": self.name,
            "class_name": self.__class__.__name__,
            "calling_id": self.calling_id,
        }

    def halt(self):
        self.continue_running = False

    @abstractmethod
    def call(self, *args, **kwargs):
        """
        这是同步调用的主入口，默认须实现该方法。
        """
        raise NotImplementedError("子类必须实现 call 方法")

    async def async_call(self, *args, **kwargs):
        """
        默认的异步调用，通过多线程实现。
        请注意，这会制造出大量线程，并不是最佳的性能优化方案。
        虽然不适合大规模部署，但这一方案可以在无需额外开发的情况下支持在异步环境中调用，快速验证业务逻辑。
        """
        loop = asyncio.get_running_loop()
        for block in await self.run_in_executor(self.call, *args, **kwargs):
            yield block

    async def run_in_executor(self, sync_function: Callable, *args, **kwargs):
        loop = asyncio.get_running_loop()
        func = partial(sync_function, *args, **kwargs)
        return await loop.run_in_executor(self.executor, func)

    def bind_consumer(self, runnable, binding_map: Dict=None, dynamic: bool=False):
        """
        传递自身的 handlers 给下游 runnable。
        """
        super().bind_consumer(runnable, binding_map=binding_map, dynamic=dynamic)
        for handler in self.handlers:
            if handler not in runnable.handlers:
                runnable.handlers.append(handler)



from ..core.runnable.agent.data import PandasAgent, MatplotAgent, PythonAgent
from ..community.dashscope import Text2ImageWanx, CosplayWanx
from ..community.zhipu import CogView, CogVideoX, WebSearch

__all__ = [
    "PandasAgent",
    "MatplotAgent",
    "PythonAgent",
    "Text2ImageWanx",
    "CosplayWanx",
    "WebSearch",
]



from typing import Set, Union
import uuid
import re

from ....io import log, EventBlock
from ....config import get_env
from ..base import Runnable

class Team(Runnable):
    """
    管理多个 ChatAgent 实例，以及他们之间的共享资源、事件收集。
    """
    def __init__(
        self,
        agents: Set[Union["ChatAgent"]]=None,
        default_agent: Union["ChatAgent"]=None,
        store: dict=None,
        chat_learn_folder: str=None,
        name: str=None,
        description: str=None,
        chunk_types: list=None,
        other_types: list=None,
        **kwargs
    ):
        super().__init__(
            **kwargs
        )

        self.agents = agents if agents else set()
        self.agents_thread_ids = {}
        self.default_agent = default_agent or next(iter(self.agents), None)
        self.chat_learn_folder = chat_learn_folder or get_env("ILLUFLY_CHAT_LEARN")

        self.store = store or {}
        self.chunk_types = chunk_types or ["chunk", "tool_resp_chunk", "text", "tool_resp_text"]
        self.other_types = other_types or ["warn", "error", "info"]

        self.handlers.append(self.collect_event)
        if self.block_processor is None:
            self.block_processor = self.event_stream

        self.last_history_id = None
        self.create_new_history()

    def __repr__(self):
        return f"Team(name={self.name}, agents={self.names}, folder={self.folders})"

    @property
    def names(self):
        return [agent.name for agent in self.agents]

    @property
    def folders(self):
        return [self.chat_learn_folder]

    def create_new_history(self):
        self.last_history_id = str(uuid.uuid1())
        self.store[self.last_history_id] = {
            "threads": set({}),
            "callings": {}
        }

    def call(self, prompt: str, **kwargs):
        """
        根据 prompt 中包含的 @agent_name 名称调用团队成员，如果未指定就调用默认成员。
        """
        names = self.fetch_agent_names(prompt)
        for name in names:
            prompt = re.sub(rf"(^|\s)@{name}\s", " ", prompt)
        prompt = prompt.strip()
        for agent in self.get_agents(names):
            yield from agent.call(prompt, **kwargs)

    def fetch_agent_names(self, prompt: str):
        """
        返回agent名称列表，如果列表为空就返回self.agents中的第一个
        """
        agent_names = re.findall(r"(?:^|\s)@(\w+)(?=\s)", prompt)
        if agent_names:
            return agent_names
        else:
            return [next(iter(self.agents)).name] if self.agents else []

    def get_agents(self, names: str):
        """
        根据名称返回智能体
        """
        return [agent for agent in self.agents if agent.name in names]

    def hire(self, *agents: Union["ChatAgent"]):
        """
        雇佣一个智能体。
        如果名字相同会保留已有实例，不会重复创建。
        """
        for agent in agents:
            agent.team = self
            self.agents.add(agent)

    def fire(self, *agents: Union["ChatAgent"]):
        """
        解雇一个智能体。
        """
        for agent in agents:
            agent.team = None
            self.agents.discard(agent)

    @property
    def event_stream(self):
        """
        生成适合于 Web 处理的 SSE 事件流格式的数据。
        """
        valid_block_types = self.chunk_types + self.other_types
        def _event_stream(block, verbose: bool=False, **kwargs):
            if isinstance(block, EventBlock) and block.block_type in valid_block_types:
                return f"event: {block.block_type}\ndata: {block.text}\n"
            else:
                return ""

        return _event_stream

    @property
    def collect_event(self):
        """
        收集事件到 store 中。

        每个 EventBlock 都包含 runnable_info 属性，其中包含 thread_id 和 calling_id。
        只有最初发起调用的 Runnable 才创建为一个 calling_id，在嵌套调用时，需要将 calling_id 传递给被调用的 Runnable。

        事件流的层次结构分为：
        - history_id 标记一个完整的对话流历史
        - calling_id 标记一次调用
        - agent_name 标记一次调用中，由哪个智能体发起
        - thread_id 如果智能体包含连续记忆，则标记连续记忆的 thread_id
        - segments 将 chunk 类事件收集到一起，形成完整段落
        """
        def _collect(event, **kwargs):
            if event.runnable_info.get("thread_id", None):
                thread = (
                    event.runnable_info["name"],
                    event.runnable_info["thread_id"],
                )
                self.store[self.last_history_id]["threads"].add(thread)

            if self.last_history_id not in self.store:
                self.store[self.last_history_id]["callings"] = {}

            calling_id = event.runnable_info["calling_id"]
            if calling_id not in self.store[self.last_history_id]["callings"]:
                self.store[self.last_history_id]["callings"][calling_id] = {
                    "agent_name": event.runnable_info["name"],
                    "input": "",
                    "output": "",
                    "segments": {},
                    "other_events": []
                }

            node = self.store[self.last_history_id]["callings"][calling_id]

            if event.block_type == "user":
                node["input"] = event.text
            elif event.block_type in self.chunk_types:
                node["segments"][event.content_id] = node["segments"].get(event.content_id, "") + event.text
            elif event.block_type == "final_text":
                node["output"] = event.text
            elif event.block_type in self.other_types or self.other_types == "__all__":
                node["other_events"].append(event.json)
            else:
                pass

        return _collect


import re
import uuid
from typing import List, Union, Generator, AsyncGenerator, Callable
from .....io import EventBlock, NewLineBlock
from .....utils import minify_text, filter_kwargs, raise_invalid_params
from ...selector import Selector, End
from ..base import BaseAgent

class FlowAgent(BaseAgent):
    @classmethod
    def allowed_params(cls):
        return {
            "max_steps": "最大步骤数",
            **BaseAgent.allowed_params(),
        }

    def __init__(self, *agents, max_steps: int=None, **kwargs):
        """
        初始化 FlowAgent

        agents 是一个字典列表，键值是流程中的名称，值是一个 BaseAgent 或 Selector 实例。
        也可以直接传入一个 BaseAgent 列表，由初始化函数自动生成相应的键名。
        """
        raise_invalid_params(kwargs, self.allowed_params())
        super().__init__(**filter_kwargs(kwargs, self.allowed_params()))

        self.max_steps = max_steps or 20

        self.thread_id = None
        self.task = None
        self.agents = []
        self.agents_index = {}

        for i, agent in enumerate(agents):
            _node_name = None
            _agent = None
            if isinstance(agent, dict):
                _node_name = list(agent.keys())[0]
                _agent = self.convert_base_agent(list(agent.values())[0])
            else:
                _agent = self.convert_base_agent(agent)
                if _agent.name in self.agents_index:
                    _node_name = f"{_agent.name}-{len(self.agents_index)}"
                else:
                    _node_name = _agent.name

            self.agents.append((_node_name, _agent))
            self.agents_index[_node_name] = (i, _agent)

    def __repr__(self):
        return f"FlowAgent({[agent[0] for agent in self.agents]})"

    def convert_base_agent(self, agent):
        if isinstance(agent, (BaseAgent, Selector, End)):
            return agent
        elif isinstance(agent, Callable):
            return BaseAgent(name=agent.__name__, func=agent)
        else:
            raise ValueError("agent must be a BaseAgent, Selector or Callable")

    def get_agent(self, name):
        return self.agents_index.get(name, (None, None))[1]

    def begin_call(self, args):
        """
        开始执行的回调方法。
        """
        self.thread_id = str(uuid.uuid1())
        self.task = args[0] if args else None
        return args

    def end_call(self):
        """
        结束执行的回调方法。
        """
        pass

    def is_end(self, agent):
        if isinstance(agent, str):
            return agent.lower() == "__end__"
        elif isinstance(agent, BaseAgent):
            return agent.name.lower() == "__end__"

        raise ValueError("agent must be a str or BaseAgent", agent)

    def call(self, *args, **kwargs):
        """
        执行智能体管道。
        """

        # 初始化当前节点信息
        (current_node_name, current_agent) = self.agents[0]
        current_index = 0

        # 初始化调用参数
        current_kwargs = kwargs

        # 初始化总步数
        steps_count = 1

        # 开始回调
        current_args = self.begin_call(args)

        while(steps_count < self.max_steps):
            # 从 current_node_name 获得 agent 和 当前 index
            (current_index, selected_agent) = self.agents_index.get(current_node_name, (None, None))
            if current_index is None:
                yield EventBlock("warn", f"节点 {current_node_name} 不存在")
                break

            # 如果 selected_agent 是一个选择器
            if isinstance(selected_agent, Selector):
                # 如果是选择器，就执行一次 select 方法
                selected_agent.select()
                # 更新当前节点的名称 current_node_name
                if isinstance(selected_agent.selected, str):
                    current_node_name = selected_agent.selected

                    # 如果已经到了 __End__  节点，就退出
                    if current_node_name.lower() == "__end__":
                        yield EventBlock("info", f"到达 __End__ 节点，结束")
                        break

                    (current_index, selected_agent) = self.agents_index.get(current_node_name, (None, None))
                    if current_index is None:
                        yield EventBlock("warn", f"节点 {current_node_name} 不存在")
                        break
                else:
                    current_node_name = selected_agent.selected.name
            elif isinstance(selected_agent, End):
                yield EventBlock("info", f"到达 __End__ 节点，结束")
                break

            # 广播节点信息给 handlers
            info = self._get_node_info(current_index + 1, current_node_name)
            yield EventBlock("node", info)

            # 执行当前节点的 call 方法
            call_resp = selected_agent.selected.call(*current_args, **current_kwargs)
            if isinstance(call_resp, Generator):
                yield from call_resp

            # 分析当前节点的输出
            if selected_agent.selected.last_output:
                # 如果节点已经有了最终的输出，就保存到 FlowAgent 的 last_output 属性中
                self._last_output = selected_agent.selected.last_output

            # 构造下一次调用的参数
            current_args = [selected_agent.selected]

            if (current_index + 1) >= len(self.agents):
                # 如果已经超出最后一个节点，就结束
                yield EventBlock("warn", f"超出最后一个节点，结束")
                break
            else:
                # 否则继续处理下一个节点
                (current_node_name, current_agent) = self.agents[current_index + 1]

            # 总步数递增
            steps_count += 1

        # 结束回调
        self.end_call()

        yield EventBlock("info", f"执行完毕，所有节点运行 {steps_count} 步")

    async def async_call(self, *args, **kwargs):
        """
        执行智能体管道。
        """

        # 初始化当前节点信息
        (current_node_name, current_agent) = self.agents[0]
        current_index = 0

        # 初始化调用参数
        current_kwargs = kwargs

        # 初始化总步数
        steps_count = 1

        # 开始回调
        current_args = self.begin_call(args)

        while(steps_count < self.max_steps):
            # 从 current_node_name 获得 agent 和 当前 index
            (current_index, selected_agent) = self.agents_index.get(current_node_name, (None, None))
            if current_index is None:
                yield EventBlock("warn", f"节点 {current_node_name} 不存在")
                break

            # 如果 selected_agent 是一个选择器
            if isinstance(selected_agent, Selector):
                # 如果是选择器，就执行一次 select 方法
                selected_agent.select()
                # 更新当前节点的名称 current_node_name
                if isinstance(selected_agent.selected, str):
                    current_node_name = selected_agent.selected

                    # 如果已经到了 __End__  节点，就退出
                    if current_node_name.lower() == "__end__":
                        yield EventBlock("info", f"到达 __End__ 节点，结束")
                        break

                    (current_index, selected_agent) = self.agents_index.get(current_node_name, (None, None))
                    if current_index is None:
                        yield EventBlock("warn", f"节点 {current_node_name} 不存在")
                        break
                else:
                    current_node_name = selected_agent.selected.name
            elif isinstance(selected_agent, End):
                yield EventBlock("info", f"到达 __End__ 节点，结束")
                break

            # 广播节点信息给 handlers
            info = self._get_node_info(current_index + 1, current_node_name)
            yield EventBlock("node", info)

            # 执行当前节点的 call 方法
            call_resp = selected_agent.selected.async_call(*current_args, **current_kwargs)
            if isinstance(call_resp, AsyncGenerator):
                async for block in call_resp:
                    yield block
            else:
                # 如果 call_resp 不是异步生成器，直接 await
                result = await call_resp
                yield result

            # 分析当前节点的输出
            if selected_agent.selected.last_output:
                # 如果节点已经有了最终的输出，就保存到 FlowAgent 的 last_output 属性中
                self._last_output = selected_agent.selected.last_output

            # 构造下一次调用的参数
            current_args = [selected_agent.selected]

            if (current_index + 1) >= len(self.agents):
                # 如果已经超出最后一个节点，就结束
                yield EventBlock("warn", f"超出最后一个节点，结束")
                break
            else:
                # 否则继续处理下一个节点
                (current_node_name, current_agent) = self.agents[current_index + 1]

            # 总步数递增
            steps_count += 1

        # 结束回调
        self.end_call()

        yield EventBlock("info", f"执行完毕，所有节点运行 {steps_count} 步")


    def _get_node_info(self, index, node_name):
        return f">>> flow-node-{index}: {node_name}"




import json
import os
import requests
import uuid

from ...types import BaseAgent
from ...utils import raise_invalid_params
from ..http import (
    EventBlock,
    send_request,
    ZHIPUAI_API_TOOLS_BASE_URL,
)

class WebSearch(BaseAgent):
    """
    [API](https://open.bigmodel.cn/dev/api/search-tool/web-search-pro)
    """
    @classmethod
    def allowed_params(cls):
        return {
            "model": "模型名称",
            "api_key": "API_KEY",
            **BaseAgent.allowed_params()
        }

    def __init__(self, tool: str=None, api_key: str=None, **kwargs):
        raise_invalid_params(kwargs, self.__class__.allowed_params())

        super().__init__(threads_group="WebSearch", **kwargs)

        self.description = "搜索互联网内容。"
        self.tool_params = {
            "prompt": "请给出互联网搜索的内容关键字",
        }

        self.tool = tool or "web-search-pro"
        self.api_key = api_key or os.getenv("ZHIPUAI_API_KEY")

    def call(
        self, 
        prompt: str=None,
        **kwargs
    ):
        self._last_output = ""

        data = {
            "request_id": str(uuid.uuid1()),
            "tool": self.tool,
            "stream": True,
            "messages": [{"role": "user", "content": prompt}]
        }

        resp = requests.post(
            ZHIPUAI_API_TOOLS_BASE_URL,
            json=data,
            headers={'Authorization': self.api_key},
            timeout=300
        )

        resp_content = resp.content.decode()
        for line in resp_content.splitlines():
            if line.startswith("data: "):
                data = line[len("data: "):]
                if data == "[DONE]":
                    break
                content = json.loads(data)
                if content.get("model", None) == "web-search-pro":
                    choices = content.get("choices", [])
                    for choice in choices:
                        for item in choice.get("delta", {}).get("tool_calls", []):
                            indent_type = item.get("type", None)
                            if indent_type == "search_intent":
                                yield EventBlock("WEB_SEARCH_INTENT", item.get("search_intent", ""))
                            elif indent_type == "search_result":
                                output = item.get("search_result", "")
                                if output:
                                    text = "\n\n".join([
                                        f'## [{r["title"]}]({r["link"]})\n{r["content"]}'
                                        for r
                                        in output
                                    ])
                                    yield EventBlock("TEXT", text)
                                    self._last_output += text




import matplotlib.pyplot as plt

from .python_code import PythonAgent

class MatplotAgent(PythonAgent):
    """
    使用 matplotlib 绘制图表的 Agent。
    """

    def __init__(self, *args, template_id: str=None, **kwargs):
        self.template_id = template_id or "CODE/Matplot"

        # 生成作为工具被使用时的功能描述
        _tool_params = kwargs.pop("tool_params", {
            "question": "细致描述图表绘制任务的需求描述",
        })
        super().__init__(*args, tool_params=_tool_params, **kwargs)



import copy
import pandas as pd
import asyncio

from typing import Union, List, Dict, Any, Callable, Generator, AsyncGenerator

from .tool_ability import ToolAbility
from ..base import Runnable
from ....io import EventBlock
from ....utils import filter_kwargs, raise_invalid_params

class BaseAgent(Runnable, ToolAbility):
    """    
    基于 BaseAgent 子类可以实现多智能体协作。

    什么时候直接从 BaseAgent 继承？
    - 需要 Runnable 基类的能力的同时
    - 还需要作为工具被使用
    - 例如，多模态模型

    什么时候转而使用ChatAgent？
    - 需要管理记忆、知识、数据等上下文
    - 例如，对话模型
    """
    @classmethod
    def allowed_params(cls):
        """
        返回当前可用的参数列表。
        """
        return {
            "func": "用于自定义工具的同步执行函数",
            "async_func": "用于自定义工具的异步执行函数",
            **Runnable.allowed_params(),
            **ToolAbility.allowed_params()
        }

    def __init__(
        self,
        func: Callable=None,
        async_func: Callable=None,
        **kwargs
    ):
        raise_invalid_params(kwargs, self.__class__.allowed_params())

        Runnable.__init__(self, **filter_kwargs(kwargs, Runnable.allowed_params()))

        name = kwargs.pop("name", self.name)
        description = kwargs.pop("description", None)

        _func = func or async_func
        if _func:
            self.name = kwargs.get("name", _func.__name__)
            name = self.name
            description = _func.__doc__ if _func.__doc__ else description

        ToolAbility.__init__(
            self,
            func=func,
            async_func=async_func,
            description=description,
            **filter_kwargs(kwargs, ToolAbility.allowed_params())
        )

    @property
    def runnable_info(self):
        info = super().runnable_info
        info.update({
            "agent_name": self.name,
            "agent_description": self.description,
        })
        return info

    def call(self, *args, **kwargs):
        if not isinstance(self.func, Callable):
            raise ValueError("func must be a callable")

        try:
            resp = self.func(*args, **kwargs)

            if isinstance(resp, Generator):
                for block in resp:
                    if isinstance(block, EventBlock) and block.block_type == "final_text":
                        self._last_output = block.content
                    yield block
            else:
                self._last_output = resp
                yield EventBlock("text", resp)

        except Exception as e:
            print(e)
            resp = str(e)
            yield EventBlock("text", resp)

    async def async_call(self, *args, **kwargs):
        try:
            if self.async_func:
                # 使用 asyncio.iscoroutinefunction 判断是否为协程函数
                if asyncio.iscoroutinefunction(self.async_func):
                    resp = await self.async_func(*args, **kwargs)
                else:
                    resp = await self.run_in_executor(self.async_func, *args, **kwargs)
            else:
                resp = await self.run_in_executor(self.func, *args, **kwargs)

            if isinstance(resp, Generator):
                for block in resp:
                    if isinstance(block, EventBlock) and block.block_type == "final_text":
                        self._last_output = block.content
                    yield block
            elif isinstance(resp, AsyncGenerator):
                async for block in resp:
                    if isinstance(block, EventBlock) and block.block_type == "final_text":
                        self._last_output = block.content
                    yield block
            else:
                self._last_output = resp
                yield EventBlock("text", resp)

        except Exception as e:
            resp = str(e)
            yield EventBlock("text", resp)



from typing import Union, List, Dict, Any, Callable, Tuple
import copy

class BindingManager:
    """
    BindingManager 基类，用于管理绑定相关的功能。

    关于动态绑定：有时仅希望短暂维持绑定关系，例如在调用函数中临时建立的绑定关系，希望每次重置。
    这与实例声明时希望长期建立的绑定关系不同，称为动态绑定。

    绑定机制的限制：
    1. 不能重复绑定，重复绑定将被忽略。
    2. 不能扩散传播 None 的 provider 键值。
    """
    @classmethod
    def allowed_params(cls):
        """
        返回当前可用的参数列表。
        """
        return {
            "providers": "实例的 consumer_dict 属性由 providers 列表中每个 Runnable 的 provider_dict 属性提供",
            "consumers": "实例的 provider_dict 属性将被 consumers 列表中每个 Runnable 引用",
            "dynamic_providers": "如果实例在不同周期中重复使用，可能会希望先在绑定前先清除旧的绑定，此时就应该使用动态绑定，即执行 bind_provider 时提供 dynamic=True 参数",
            "lazy_binding_map": "有时你无法确定被哪个对象绑定，但能确定绑定映射，此时就可以使用 lazy_binding_map 参数，在绑定时由对方根据该参数进行绑定",
        }

    def __init__(
        self,
        providers: List[Tuple[Any, Dict]]=None,
        consumers: List[Tuple[Any, Dict]]=None,
        dynamic_providers: List[Tuple[Any, Dict]]=None,
        lazy_binding_map: Dict=None
    ):
        """
        :param binding: 绑定的 (runnable, binding_map)，字典结构，或 Runnable 实例的列表
        """
        self.providers = providers or []
        self.consumers = consumers or []
        self.dynamic_providers = dynamic_providers or []

        # lazy_binding_map 可用于被绑定时默认采纳的 binding_map
        self.lazy_binding_map = lazy_binding_map or {}

    @property
    def provider_dict(self):
        """
        用于被其他 runnable 对象绑定的变量。
        实现绑定变量的传递：将导入的变量合并后作为导出的变量。

        需要注意的是，值为 None 的绑定变量不会被传递；
        但仅针对明确的 None 值，真是因为真的希望传递空值，所以 ""、[] 等非 None 空值仍然会被传递。
        """
        return {k: v for k, v in self.consumer_dict.items() if v is not None}

    def clear_dynamic_providers(self):
        """
        清除所有动态绑定的 providers
        """
        self.dynamic_providers = []

    def _convert_binding(self, runnable: "Runnable"=None, binding_map: Dict=None):
        """
        将 binding 转换为 (runnable, binding_map) 标准形式。
        """
        if not runnable and not binding_map:
            raise ValueError("runnable and binding_map cannot be both empty")

        # 这里做一个语法糖转换，如果传入的是一个字典，则表示 binding_map，runnable 为 None        
        if isinstance(runnable, dict) and binding_map is None:
            binding_map = copy.deepcopy(runnable)
            runnable = None

        binding_map = binding_map or self.lazy_binding_map
        runnable = runnable or PassthroughBinding(binding_map=binding_map)

        if not isinstance(runnable, BindingManager):
            raise ValueError("runnable must be a Runnable instance", runnable)
        if not isinstance(binding_map, dict):
            raise ValueError("binding_map must be a dictionary", binding_map)

        return (runnable, binding_map)

    def bind_provider(self, runnable: "Runnable"=None, binding_map: Dict=None, dynamic: bool=False):
        """
        绑定其他 providers 以便使用其输出字典。

        可以按 binding 中的映射规则绑定到 Runnable 实例，然后动态获取值；
        也可以直接从指定 dict 结构中获取值。

        动态绑定：
        如果 dynamic 为 True，则清除所有动态绑定的 providers，并添加新的动态绑定。
        """
        (provider_runnable, binding_map) = self._convert_binding(runnable, binding_map)

        if not isinstance(provider_runnable, BindingManager):
            raise ValueError("provider runnable must be a Runnable instance", provider_runnable)

        if provider_runnable.name in [r[0].name for r in self.providers]:
            # 如果已经绑定则忽略，不重复绑定
            return

        if dynamic:
            self.clear_dynamic_providers()
            self.dynamic_providers.append((provider_runnable, binding_map))
        else:
            self.providers.append((provider_runnable, binding_map))
            provider_runnable.consumers.append((self, binding_map))

    def bind_consumer(self, runnable: "Runnable", binding_map: Dict=None, dynamic: bool=False):
        """
        将自身绑定给 consumers 以便将输出字典提供其使用。

        与 bind_provider 不同，不存在将自己的输出字典传递给一个字典的情况，因此被绑定的消费者 runnable 不能为空。
        """
        if not isinstance(runnable, BindingManager):
            raise ValueError("consumer runnable must be a Runnable instance", runnable)
        runnable.bind_provider(self, binding_map, dynamic=dynamic)

    @property
    def consumer_dict(self):
        """
        获取所有绑定的 runnable 的 provider_dict 变量。

        如果设定了 binding_map，则将 runnable 中的变量通过 binding_map 进行映射。

        规则1 如果被绑定 Runnable 的导出变量没有被使用，则按同名导入。
        规则2 如果 binding_map 中映射的值是 str，则使用 provider_dict 中的变量进行映射。
        规则3 如果 binding_map 中映射的值是 Callable，则执行映射函数。
        规则4 如果 binding_map 中映射的值是 None 则放弃映射。

        使用函数扩展时，不会覆盖函数中包含的键值，这实际上提供了 **1:N** 映射的可能性。
        """

        consumer_dict = {}

        for runnable, binding_map in self.providers + self.dynamic_providers:
            vars = runnable.provider_dict
            for k, v in vars.items():
                mapping_index = {iv: ik for ik, iv in binding_map.items() if not isinstance(iv, Callable)}
                if k in mapping_index:
                    # 实现规则 2
                    consumer_dict[mapping_index[k]] = v
                else:
                    # 实现规则 1
                    consumer_dict[k] = v
            for k, v in binding_map.items():
                if isinstance(v, Callable):
                    # 实现规则 3
                    if v.__code__.co_argcount == 0:
                        consumer_dict[k] = v()
                    else:
                        consumer_dict[k] = v(vars)
                elif v is None:
                    # 实现规则 4
                    consumer_dict[k] = None

        return {k:v for k,v in consumer_dict.items() if v is not None}

    @property
    def provider_tree(self):
        """
        从当前 runnable 开始，获取所有上游绑定的 provider 及其祖先。
        """
        tree = {"consumer": self, "provider_tree": []}
        for provider, binding_map in self.providers:
            tree["provider_tree"].append({
                "provider": provider,
                "binding_map": binding_map,
                "provider_tree": provider.provider_tree
            })
        return tree
    
    @property
    def consumer_tree(self):
        """
        从当前 runnable 开始，获取所有下游绑定的 consumer 及其子孙。
        """
        tree = {"provider": self, "consumer_tree": []}
        for consumer, binding_map in self.consumers:
            tree["consumer_tree"].append({
                "consumer": consumer,
                "binding_map": binding_map,
                "consumer_tree": consumer.consumer_tree
            })
        return tree

class PassthroughBinding(BindingManager):
    """
    透传字典结构到 provider_dict
    """
    def __init__(self, binding_map: Dict=None):
        super().__init__()
        self.name = f"PassthroughBinding-{id(self)}"
        self.binding_map = binding_map or {}

    @property
    def provider_dict(self):
        return {k: v for k, v in self.binding_map.items() if v is not None}



from .embeddings import TextEmbeddings
from .qwen import ChatQwen, ChatQwenVL
from .wanx import (
    Text2ImageWanx,
    CosplayWanx,
    RepaintWanx,
    StyleRepaintWanx,
    BackgroundWanx,
    AnyTextWanx,
    SketchWanx,
    WordArt,
    WordArtSemantic
)



from typing import List
from ....io import EventBlock
from ....utils import raise_invalid_params
from ..base import Runnable
from ..embeddings import BaseEmbeddings

class VectorDB(Runnable):
    """
    向量数据库。
    """
    @classmethod
    def allowed_params(cls):
        return {
            "embeddings": "用于相似度计算的 embeddings 对象",
            "top_k": "返回结果的条数，默认 5",
            **Runnable.allowed_params()
        }

    def __init__(self, embeddings: BaseEmbeddings, top_k: int=None, **kwargs):
        raise_invalid_params(kwargs, self.__class__.allowed_params())

        if not isinstance(embeddings, BaseEmbeddings):
            raise ValueError(f"{embeddings} must be an instance of BaseEmbeddings")
        if embeddings.dim is None:
            raise ValueError(f"{embeddings} dim must be set")

        super().__init__(**kwargs)

        # 必须提供一个向量模型
        self.embeddings = embeddings
        self.dim = embeddings.dim
        self.top_k = top_k

        # 向量数据库所使用的文档来源
        self.sources = []

        # 向量数据库中备查的文档，包括了原始文档、元数据，以及由向量模型转换的向量索引
        self.documents = []

    def query(self, vector: List[List[float]], top_k: int=None, **kwargs):
        pass

    def call(self, text: str, top_k: int=None, **kwargs):
        self._last_output = self.query(text, top_k or self.top_k or 5, **kwargs)
        yield EventBlock("info", f"查询到{len(self._last_output)}条结果")

    def load_text(self, text: str):
        pass

    def load(self, dir: str):
        pass



from .block import EventBlock, EndBlock, NewLineBlock, ResponseBlock

from .handlers import log, alog, usage, async_usage
from .block_process import event_stream

from ..core.history import BaseHistory, LocalFileHistory, InMemoryHistory

__all__ = [
    "EventBlock",
    "EndBlock",
    "NewLineBlock",
    "ResponseBlock",
    "log",
    "alog",
    "event_stream",
    "usage",
    "async_usage",
    "BaseHistory",
    "LocalFileHistory",
    "InMemoryHistory",
]



import inspect
import json
import textwrap
from decimal import Decimal

from typing import Any, Callable, Dict, get_origin, get_args, Union, Tuple, Set

class ToolAbility:
    @classmethod
    def allowed_params(cls):
        """
        返回当前可用的参数列表。
        """
        return {
            "func": "用于自定义工具的同步执行函数",
            "async_func": "用于自定义工具的异步执行函数",
            "name": "工具名称",
            "description": "工具描述",
            "tool_params": "工具参数",
        }

    def __init__(
        self,
        *,
        func: Callable = None,
        async_func: Callable = None,
        name: str = None,
        description: str = None,
        tool_params: Dict[str, Any] = None
    ):
        self.func = func
        self.async_func = async_func
        self.description = description or "我还没有工具描述"
        self.tool_params = tool_params or {}

        if name:
            self.name = name

    def get_default_parameters(self):
        _parameters = {
            "type": "object",
            "properties": {},
            "required": []
        }
        sig = inspect.signature(self.func or self.async_func or self.call)
        for name, param in sig.parameters.items():
            param_value = param.default if param.default is not inspect.Parameter.empty else ""
            
            # 根据参数的注解类型设置 JSON 兼容类型
            param_type = self._get_json_type(param.annotation)
            
            _parameters["properties"][name] = {
                "type": param_type,
                "description": str(param_value)
            }
            if param.default is inspect.Parameter.empty:
                _parameters["required"].append(name)
        return _parameters

    def _get_json_type(self, annotation):
        origin = get_origin(annotation)
        args = get_args(annotation)

        if origin is Union:
            # 处理 Union 类型，返回所有可能的类型
            return [self._get_json_type(arg) for arg in args]
        elif origin is list or annotation is list:
            return "array"
        elif origin is tuple or annotation is tuple:
            # 假设 Tuple 是一个固定长度的数组
            return {
                "type": "array",
                "items": [self._get_json_type(arg) for arg in args] if args else "string"
            }
        elif origin is set or annotation is set:
            # Set 可以被视为一个不允许重复元素的数组
            return {
                "type": "array",
                "uniqueItems": True,
                "items": self._get_json_type(args[0]) if args else "string"
            }
        elif origin is dict or annotation is dict:
            return "object"
        elif annotation is int:
            return "integer"
        elif annotation is float:
            return "number"
        elif annotation is bool:
            return "boolean"
        elif annotation is str:
            return "string"
        else:
            return "string"  # 默认类型

    @property
    def parameters(self) -> str:
        default_params = self.get_default_parameters()
        final_params = {'type': default_params["type"], 'properties': {}, 'required': []}
        if not self.tool_params:
            return default_params
        else:
            for prop_name, prop_dict in default_params["properties"].items():
                if prop_name in self.tool_params.keys():
                    final_params["properties"][prop_name] = prop_dict
                    final_params["properties"][prop_name]["description"] = self.tool_params[prop_name]
            final_params["required"] = [k for k in default_params["required"] if k in final_params["properties"].keys()]
            return final_params

    @property
    def tool_desc(self) -> Dict[str, Any]:
        """
        当作为工具时的自描述信息。

        内容参考了 openai 的工具描述规格生成。
        """
        return {
            "type": "function",
            "function": {
                "name": self.name,
                "description": self.description,
                "parameters": self.parameters,
            }
        }

    def parse_arguments(self, arguments: str) -> Dict[str, Any]:
        """
        解析 arguments 字符串并根据 self.parameters 中的类型信息转换为相应的 Python 类型。
        """
        try:
            parsed_arguments = json.loads(arguments)
        except json.JSONDecodeError as e:
            raise ValueError(f"Invalid JSON input: {e}")

        parameter_types = self.parameters.get("properties", {})

        if isinstance(parsed_arguments, list):
            return parsed_arguments
        elif isinstance(parsed_arguments, dict):
            for key, value in parsed_arguments.items():
                expected_type = parameter_types.get(key, {}).get("type", "string")
                
                try:
                    if expected_type == "integer":
                        parsed_arguments[key] = int(value)
                    elif expected_type == "number":
                        parsed_arguments[key] = float(value)
                    elif expected_type == "boolean":
                        if isinstance(value, str):
                            parsed_arguments[key] = value.lower() == "true"
                        else:
                            parsed_arguments[key] = bool(value)
                    elif expected_type == "array":
                        if "uniqueItems" in parameter_types.get(key, {}):
                            parsed_arguments[key] = set(value)
                        elif "items" in parameter_types.get(key, {}):
                            parsed_arguments[key] = tuple(value)
                        else:
                            parsed_arguments[key] = list(value)
                    elif expected_type == "null":
                        parsed_arguments[key] = None
                    elif expected_type == "decimal":
                        parsed_arguments[key] = Decimal(value)
                    else:
                        parsed_arguments[key] = value
                except (ValueError, TypeError) as e:
                    raise ValueError(f"Error converting {key}: {e}")
            return parsed_arguments
        else:
            raise ValueError("Parsed arguments must be a list or a dictionary")







from typing import Dict, Any, List, Union
from ..utils import minify_text
import json
import time
import random

class IDGenerator:
    def __init__(self):
        self.counter = 0

    def create_id(self):
        while True:
            timestamp = str(int(time.time()))[-4:]
            random_number = f'{random.randint(0, 999):03}'
            counter_str = f'{self.counter:03}'
            yield f'{timestamp}-{random_number}-{counter_str}'
            self.counter = 0 if self.counter == 999 else self.counter + 1

id_generator = IDGenerator()
id_gen = id_generator.create_id()

class Document():
    def __init__(self, text: str=None, index: str=None, meta: Dict[str, Any] = None):
        self.text = text or ""
        self.index = index or ""
        self.meta = {**(meta or {})}
        if 'id' not in self.meta:
            self.meta['id'] = next(id_gen)

    def __repr__(self):
        meta_converted = [k for k, _v in self.meta.items()]
        return f"Document(text=\"{minify_text(self.text)}\", meta={meta_converted})"

    def __str__(self):
        return self.text
    
    def __len__(self):
        return len(self.text)

def convert_to_documents_list(docs: Union[str, List[str], List[Document]]):
    if isinstance(docs, str):
        docs = [docs]

    if not isinstance(docs, list):
        raise ValueError("docs 必须是字符串或 Document 类型列表，但实际为: {type(docs)}")
    
    docs_list = []
    for doc in docs:
        if isinstance(doc, str):
            doc = Document(doc, meta={'source': '__rerank__'})
        if not isinstance(doc, Document):
            raise ValueError("docs 必须是字符串或 Document 类型列表，但实际为: {type(docs)}")

        docs_list.append(doc)

    return docs_list



from .base import Runnable, BaseAgent
from .chat import ChatAgent, ChatPool
from .chat import BaseToolCalling, ToolCall, SubTask, Plans
from .retriever import Retriever
from .flow import FlowAgent, CoT, ReAct, ReWOO, PlanAndSolve
from .flow import ChatLearn
from .writer import FromOutline
from .data import PandasAgent, MatplotAgent, PythonAgent
